#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  i18n.py
#  
#  Copyright 2018 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "TranslationManager",
]

import glob
import re

from .util import ActionDispatcher

# TODO: add test cases for translations
class TranslationManager(ActionDispatcher):
    """
    Manages sets of translation files in multiple languages.
    
    This Translation System uses language codes to identify languages, there is
    no requirement to follow a specific standard, but it is recommended to use
    simple 2-digit codes like ``en`` and ``de``\ , adding an underscore to
    define sub-languages like ``en_gb`` and ``en_us``\ .
    
    Whenever a new translation file is needed, it will be parsed and then cached.
    This speeds up access times and also practically eliminates load times when
    switching languages.
    
    Several events are sent by this class, see :ref:`events-i18n`\ .
    
    Most of these events are also sent as actions, these actions are described
    in the methods that cause them.
    
    There are also severale config options that determine the behaviour of this class.
    See :ref:`cfg-i18n` for more information.
    
    
    This Manager requires the :py:class:`~peng3d.resource.ResourceManager()` to
    be already initialized.
    """
    def __init__(self,peng):
        if not peng.cfg["rsrc.enable"]:
            raise RuntimeError("ResourceManager needs to be enabled to use Translations")
        elif peng.rsrcMgr is None:
            raise RuntimeError("ResourceManager needs to be initialized before TranslationManager")
        
        self.peng = peng
        
        self.lang = self.peng.cfg["i18n.lang"]
        
        self.cache = {} # dict of dicts, first lang, then domain
        
        self.peng.sendEvent("peng3d:i18n.init",{"lang":self.lang,"i18n":self})
        self.setLang(self.peng.cfg["i18n.lang"])
    
    def setLang(self,lang):
        """
        Sets the default language for all domains.
        
        For recommendations regarding the format of the language code, see
        :py:class:`TranslationManager`\ .
        
        Note that the ``lang`` parameter of both :py:meth:`translate()` and
        :py:meth:`translate_lazy()` will override this setting.
        
        Also note that the code won't be checked for existence or plausibility.
        This may cause the fallback strings to be displayed instead if the language
        does not exist.
        
        Calling this method will cause the ``setlang`` action and the
        :peng3d:event`peng3d:i18n.set_lang` event to be triggered. Note that both
        action and event will be triggered even if the language did not actually change.
        
        This method also automatically updates the :confval:`i18n.lang` config value.
        """
        self.lang = lang
        self.peng.cfg["i18n.lang"] = lang
        
        if lang not in self.cache:
            self.cache[lang]={}
        
        self.doAction("setlang")
        self.peng.sendEvent("peng3d:i18n.set_lang",{"lang":self.lang,"i18n":self})
    
    def discoverLangs(self,domain="*"):
        """
        Generates a list of languages based on files found on disk.
        
        The optional ``domain`` argument may specify a domain to use when checking
        for files. By default, all domains are checked.
        
        This internally uses the :py:mod:`glob` built-in module and the
        :confval:`i18n.lang.format` config option to find suitable filenames.
        It then applies the regex in :confval:`i18n.discover_regex` to extract the
        language code.
        """
        rsrc = self.peng.cfg["i18n.lang.format"].format(domain=domain,lang="*")
        pattern = self.peng.rsrcMgr.resourceNameToPath(rsrc,self.peng.cfg["i18n.lang.ext"])
        files = glob.iglob(pattern)
        
        langs = set()
        r = re.compile(self.peng.cfg["i18n.discover_regex"])
        
        for f in files:
            m = r.fullmatch(f)
            if m is not None:
                langs.add(m.group("lang"))
        
        return list(langs)
    
    def translate(self,key,translate=True,lang=None):
        if lang is None:
            lang = self.lang
        
        if not translate or key.count(":")!=1:
            return key
        domain,name = key.split(":")
        
        if domain not in self.cache[lang]:
            self.loadDomain(domain,lang)
        if lang not in self.cache or domain not in self.cache[lang] or name not in self.cache[lang][domain]:
            return key # would just display the key to the user, good enough to be understood
        return self.cache[lang][domain][name]
    t = translate
    def translate_lazy(self,key,translate=True,lang=None):
        return _LazyTranslator(self,key,translate,lang)
    tl = translate_lazy
    
    def loadDomain(self,domain,lang=None):
        if lang is None:
            lang = self.lang
        
        if lang not in self.cache:
            self.cache[lang][domain]={} # prevents errors if function aborts prematurely
        
        rsrc = self.peng.cfg["i18n.lang.format"].format(domain=domain,lang=lang)
        if not self.peng.rsrcMgr.resourceExists(rsrc,self.peng.cfg["i18n.lang.ext"]):
            return # prevents errors
        fname = self.peng.rsrcMgr.resourceNameToPath(rsrc,self.peng.cfg["i18n.lang.ext"])
        try:
            with open(fname,"r") as f:
                data = f.readlines()
        except Exception:
            return # prevents errors
        
        d = {}
        for line in data:
            line.strip()
            if line=="" or line.startswith("#"):
                continue
            ls = line.split("=")
            k = ls.pop(0) # first ever useful application of pop
            v = "=".join(ls).strip()
            d[k]=v
        self.cache[lang][domain]=d
        
        self.doAction("loaddomain")
    
    def __getitem__(self,key):
        return self.translate(key)

class _LazyTranslator(object):
    _dynamic=True
    def __init__(self,i18n,key,translate=True,lang=None):
        self.i18n = i18n
        self.key = key
        self.translate = translate
        self.lang = lang
    def __str__(self):
        return self.i18n.translate(self.key,self.translate,self.lang)
    def __repr__(self):
        return self.i18n.translate(self.key,self.translate,self.lang)
    def __mod__(self,other):
        try:
            return str(self)%other
        except Exception:
            return str(self)
    def format(self,*args,**kwargs):
        try:
            return str(self).format(*args,**kwargs)
        except Exception:
            return str(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  peng.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["Peng","HeadlessPeng"]

import sys

import weakref
import inspect

#from . import window, config, keybind, pyglet_patch
from . import config, world, resource, i18n

_pyglet_patched = sys.version_info.major == 2 or not world._have_pyglet

class Peng(object):
    """
    This Class should only be instantiated once per application, if you want to use multiple windows, see :py:meth:`createWindow()`\ .
    
    An Instance of this class represents the whole Engine, with all accompanying state and window/world objects.
    
    Be sure to keep your instance accessible, as it will be needed to create most other classes.
    """
    
    def __init__(self,cfg={}):
        global _pyglet_patched
        if world._have_pyglet:
            from . import pyglet_patch, keybind # Local import for compat with headless machines
        self.window = None
        
        self.pygletEventHandlers = {}
        
        self.eventHandlers = {}
        
        self.events_ignored = {}
        self.event_list = set()
        
        if cfg == {}:
            cfg = {} # To avoid bugs with default arguments
        self.cfg = config.Config(cfg,defaults=config.DEFAULT_CONFIG)
        if world._have_pyglet:
            self.keybinds = keybind.KeybindHandler(self)
        
        self.resourceMgr = None
        self.i18n = None
        
        self.t = lambda *args,**kwargs:self._t(*args,**kwargs)
        self.tl = lambda *args,**kwargs:self._tl(*args,**kwargs)
        
        self.addEventListener("peng3d:peng.exit",self.handler_exit)
        
        if not _pyglet_patched and self.cfg["pyglet.patch.patch_float2int"]:
            _pyglet_patched = True
            pyglet_patch.patch_float2int()
    
    def createWindow(self,cls=None,caption_t=None,*args,**kwargs):
        """
        createWindow(cls=window.PengWindow, *args, **kwargs)
        
        Creates a new window using the supplied ``cls``\ .
        
        If ``cls`` is not given, :py:class:`peng3d.window.PengWindow()` will be used.
        
        Any other positional or keyword arguments are passed to the class constructor.
        
        Note that this method currently does not support using multiple windows.
        
        .. todo::
           
           Implement having multiple windows.
        """
        if cls is None:
            from . import window
            cls = window.PengWindow
        if self.window is not None:
            raise RuntimeError("Window already created!")
        self.sendEvent("peng3d:window.create.pre",{"peng":self,"cls":cls})
        if caption_t is not None:
            kwargs["caption"] = "Peng3d Application"
        self.window = cls(self,*args,**kwargs)
        self.sendEvent("peng3d:window.create.post",{"peng":self,"window":self.window})
        if self.cfg["rsrc.enable"] and self.resourceMgr is None:
            self.sendEvent("peng3d:rsrc.init.pre",{"peng":self,"basepath":self.cfg["rsrc.basepath"]})
            self.resourceMgr = resource.ResourceManager(self,self.cfg["rsrc.basepath"])
            self.rsrcMgr = self.resourceMgr
            self.sendEvent("peng3d:rsrc.init.post",{"peng":self,"rsrcMgr":self.resourceMgr})
            if self.cfg["i18n.enable"] and self.i18n is None:
                self.sendEvent("peng3d:i18n.init.pre",{"peng":self})
                self.i18n = i18n.TranslationManager(self)
                self._t = self.i18n.t
                self._tl = self.i18n.tl
                self.sendEvent("peng3d:i18n.init.post",{"peng":self,"i18n":self.i18n})
        if caption_t is not None:
            self.window.set_caption(self.t(caption_t))
            def f():
                self.window.set_caption(self.t(caption_t))
            self.i18n.addAction("setlang",f)
        return self.window
    
    def run(self,evloop=None):
        """
        Runs the application main loop.
        
        This method is blocking and needs to be called from the main thread to avoid OpenGL bugs that can occur.
        
        ``evloop`` may optionally be a subclass of :py:class:`pyglet.app.base.EventLoop` to replace the default event loop.
        """
        self.sendEvent("peng3d:peng.run",{"peng":self,"window":self.window,"evloop":evloop})
        self.window.run(evloop)
        self.sendEvent("peng3d:peng.exit",{"peng":self})
    
    def sendPygletEvent(self,event_type,args,window=None):
        """
        Handles a pyglet event.
        
        This method is called by :py:meth:`PengWindow.dispatch_event()` and handles all events.
        
        See :py:meth:`registerEventHandler()` for how to listen to these events.
        
        This method should be used to send pyglet events.
        For new code, it is recommended to use :py:meth:`sendEvent()` instead.
        For "tunneling" pyglet events, use event names of the format ``pyglet:<event>``
        and for the data use ``{"args":<args as list>,"window":<window object or none>,"src":<event source>,"event_type":<event type>}``
        
        Note that you should send pyglet events only via this method, the above event will be sent automatically.
        
        Do not use this method to send custom events, use :py:meth:`sendEvent` instead.
        """
        args = list(args)
        
        self.sendEvent("pyglet:%s"%event_type,{"peng":self,"args":args,"window":window,"src":self,"event_type":event_type})
        self.sendEvent("peng3d:pyglet",{"peng":self,"args":args,"window":window,"src":self,"event_type":event_type})
        
        if event_type not in ["on_draw","on_mouse_motion"] and self.cfg["debug.events.dump"]:
            print("Event %s with args %s"%(event_type,args))
        if event_type in self.pygletEventHandlers:
            for whandler in self.pygletEventHandlers[event_type]:
                # This allows for proper collection of deleted handler methods by using weak references
                handler = whandler()
                if handler is None:
                    del self.pygletEventHandlers[event_type][self.pygletEventHandlers[event_type].index(whandler)]
                handler(*args)
    def addPygletListener(self,event_type,handler):
        """
        Registers an event handler.
        
        The specified callable handler will be called every time an event with the same ``event_type`` is encountered.
        
        All event arguments are passed as positional arguments.
        
        This method should be used to listen for pyglet events.
        For new code, it is recommended to use :py:meth:`addEventListener()` instead.
        
        See :py:meth:`handleEvent()` for information about tunneled pyglet events.
        
        For custom events, use :py:meth:`addEventListener()` instead.
        """
        if self.cfg["debug.events.register"]:
            print("Registered Event: %s Handler: %s"%(event_type,handler))
        if event_type not in self.pygletEventHandlers:
            self.pygletEventHandlers[event_type]=[]
        # Only a weak reference is kept
        if inspect.ismethod(handler):
            handler = weakref.WeakMethod(handler)
        else:
            handler = weakref.ref(handler)
        self.pygletEventHandlers[event_type].append(handler)
    
    # For compatibility
    handleEvent = sendPygletEvent
    registerEventHandler = addPygletListener
    
    def sendEvent(self,event,data=None):
        """
        Sends an event with attached data.
        
        ``event`` should be a string of format ``<namespace>:<category1>.<subcategory2>.<name>``\ .
        There may be an arbitrary amount of subcategories. Also note that this
        format is not strictly enforced, but rather recommended by convention.
        
        ``data`` may be any Python Object, but it usually is a dictionary containing relevant parameters.
        For example, most built-in events use a dictionary containing at least the ``peng`` key set to an instance of this class.
        
        If there are no handlers for the event, a corresponding message will be printed to the log file.
        To prevent spam, the maximum amount of ignored messages can be configured via :confval:`events.maxignore` and defaults to 3.
        
        If the config value :confval:`debug.events.dumpfile` is a file path, the event type will be added to an internal list and be saved to the given file during program exit.
        """
        if self.cfg["debug.events.dumpfile"]!="" and event not in self.event_list:
            self.event_list.add(event)
        if event not in self.eventHandlers:
            if event not in self.events_ignored or self.events_ignored[event]<=self.cfg["events.maxignore"]: # Prevents spamming logfile with ignored event messages
                # TODO: write to logfile
                # Needs a logging module first...
                self.events_ignored[event] = self.events_ignored.get(event,0)+1
            return
        
        for handler in self.eventHandlers[event]:
            f = handler[0]
            try:
                f(event,data)
            except Exception:
                if not handler[1]:
                    raise
                else:
                    # TODO: write to logfile
                    if self.cfg["events.removeonerror"]:
                        self.delEventListener(event,f)
    
    def addEventListener(self,event,func,raiseErrors=False):
        """
        Adds a handler to the given event.
        
        A event may have an arbitrary amount of handlers, though assigning too
        many handlers may slow down event processing.
        
        For the format of ``event``\ , see :py:meth:`sendEvent()`\ .
        
        ``func`` is the handler which will be executed with two arguments, ``event_type`` and ``data``\ , as supplied to :py:meth:`sendEvent()`\ .
        
        If ``raiseErrors`` is True, exceptions caused by the handler will be re-raised.
        Defaults to ``False``\ .
        """
        if not isinstance(event,str):
            raise TypeError("Event types must always be strings")
        
        if event not in self.eventHandlers:
            self.eventHandlers[event]=[]
        self.eventHandlers[event].append([func,raiseErrors])
    def delEventListener(self,event,func):
        """
        Removes the given handler from the given event.
        
        If the event does not exist, a :py:exc:`NameError` is thrown.
        
        If the handler has not been registered previously, also a :py:exc:`NameError` will be thrown.
        """
        if event not in self.eventHandlers:
            raise NameError("No handlers exist for event %s"%event)
        if [func,True] in self.eventHandlers[event]:
            del self.eventHandlers[event][self.eventHandlers[event].index(func)]
        elif [func,False] in self.eventHandler[event]:
            del self.eventHandlers[event][self.eventHandlers[event].index(func)]
        else:
            raise NameError("This handler is not registered for event %s"%event)
        if self.eventHandlers[event] == []:
            del self.eventHandlers[event]
    
    def handler_exit(self,event,data):
        if self.cfg["debug.events.dumpfile"]!="":
            with open(self.cfg["debug.events.dumpfile"],"w") as f:
                f.write("\n".join(sorted(list(self.event_list))))
    handler_exit.__noautodoc__ = True

class HeadlessPeng(object):
    """
    Variant of peng that should work without having pyglet installed.
    
    This class is intended for use in servers as a drop-in replacement for the normal engine class.
    
    Note that this class is only in its beginnings and should not be used yet.
    """
    def __init__(self,cfg={}):
        if "rsrc.enable" not in cfg:
            cfg["rsrc.enable"]=False
        super(HeadlessPeng,self).__init__(cfg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  container.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["Container", "ScrollableContainer", "ContainerButtonBackground"]

import collections

import pyglet
from pyglet.gl import *

from .widgets import Widget, Background, _WatchingList
from .slider import VerticalSlider
from .button import ButtonBackground
from ..layer import Layer

class ContainerButtonBackground(ButtonBackground):
    """
    Background class used to render the background of containers using a button style.
    
    Mostly identical with :py:class:`ButtonBackground` with added compatibility for containers.
    """
    change_on_press = False
    
    def getColors(self):
        bg,o,i,s,h = super(ContainerButtonBackground,self).getColors()
        i = bg
        
        return bg,o,i,s,h
    getColors.__noautodoc__ = getColors
    
    def bs_oldshadow(self,bg,o,i,s,h):
        if self.change_on_press and self.widget.pressed:
            i = s
            s,h = h,s
        elif self.change_on_press and self.widget.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
            s = [min(s[0]+6,255),min(s[1]+6,255),min(s[2]+6,255)]
        cb1 = s+s+s+s
        cb2 = h+h+h+h
        cb3 = h+h+h+h
        cb4 = s+s+s+s
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    bs_oldshadow.__noautodoc__ = getColors

class Container(Widget):
    """
    Main class of the container system.
    
    This widget may contain other widgets, limiting the childs to only draw within the defined bounds.
    Additionally, the given position will also act as a offset, making the child coordinates relative to the parent.
    
    The :py:attr:`visible` attribute may be set to control whether or not this container is visible.
    
    This Class is a subclass of :py:class:`peng3d.gui.widgets.Widget` but also exhibits part of the API of :py:class:`peng3d.gui.SubMenu`\ .
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,_skip_draw=False
                ):
        super(Container,self).__init__(name,submenu,window,peng,pos,size)
        
        self.menu = submenu
        
        self.widgets = collections.OrderedDict()
        
        self.bg = [242,241,240,255]
        self.bg_vlist = pyglet.graphics.vertex_list(4,
            "v2f",
            "c4B",
            )
        self.stencil_vlist = pyglet.graphics.vertex_list(4,
            "v2f",
            ("c4B",[0,0,0,0]*4),
            )
        self.peng.registerEventHandler("on_resize",self.on_resize)
        if not _skip_draw:
            self.on_resize(*self.submenu.size)
        
        self.batch2d = pyglet.graphics.Batch()
        
        self.visible = True
        
    def setBackground(self,bg):
        """
        Sets the background of the Container.
        
        Similar to :py:meth:`peng3d.gui.SubMenu.setBackground()`\ , but only effects the region covered by the Container.
        """
        self.bg = bg
        if isinstance(bg,list) or isinstance(bg,tuple):
            if len(bg)==3 and isinstance(bg,list):
                bg.append(255)
            self.bg_vlist.colors = bg*4
        elif bg in ["flat","gradient","oldshadow","material"]:
            self.bg = ContainerButtonBackground(self,borderstyle=bg,batch=self.batch2d)
            self.redraw()
    
    def on_resize(self,width,height):
        # Stencil/BG Vlist are updated by redraw()
        self.redraw()
    
    @property
    def clickable(self):
        if not isinstance(self.submenu,Container):
            return self.submenu.name == self.submenu.menu.activeSubMenu and self.submenu.menu.name == self.window.activeMenu and self.enabled and self.visible
        else:
            return self.submenu.clickable and self.enabled and self.visible
    @clickable.setter
    def clickable(self,value):
        self._enabled=value
        self.redraw()
    
    def addWidget(self,widget):
        """
        Adds a widget to this container.
        
        Note that trying to add the Container to itself will be ignored.
        """
        if self is widget: # Prevents being able to add the container to itself, causing a recursion loop on redraw
            return
        self.widgets[widget.name]=widget
    def getWidget(self,name):
        """
        Returns the widget with the given name.
        """
        return self.widgets[name]
    
    def draw(self):
        """
        Draws the submenu and its background.
        
        Note that this leaves the OpenGL state set to 2d drawing and may modify the scissor settings.
        """
        if not self.visible:
            # Simple visibility check, has to be tested to see if it works properly
            return
        
        if not isinstance(self.submenu,Container):
            glEnable(GL_SCISSOR_TEST)
            glScissor(*self.pos+self.size)
        
        SubMenu.draw(self)
        
        if not isinstance(self.submenu,Container):
            glDisable(GL_SCISSOR_TEST)
    
    def on_redraw(self):
        """
        Redraws the background and any child widgets.
        """
        x,y = self.pos
        sx,sy = self.size
        self.bg_vlist.vertices = [x,y, x+sx,y, x+sx,y+sy, x,y+sy]
        self.stencil_vlist.vertices = [x,y, x+sx,y, x+sx,y+sy, x,y+sy]
        if isinstance(self.bg,Background):
            if not self.bg.initialized:
                self.bg.init_bg()
                self.bg.initialized=True
            self.bg.redraw_bg()
    
    def redraw(self):
        super(Container,self).redraw()
        
        # Also schedules redraws for sub-widgets
        for widget in self.widgets.values():
            widget.redraw()
    redraw.__noautodoc__ = True
    
    def on_enter(self,old):
        """
        Dummy method defined for compatibility with :py:class:`peng3d.gui.SubMenu`, simply does nothing.
        """
        pass
    def on_exit(self,new):
        """
        Dummy method defined for compatibility with :py:class:`peng3d.gui.SubMenu`, simply does nothing.
        """
        pass

class ScrollableContainer(Container):
    """
    Subclass of :py:class:`Container` allowing for scrolling its content.
    
    The scrollbar currently is always on the right side and simply consists of a :py:class:`peng3d.gui.slider.VerticalSlider`\ .
    
    ``scrollbar_width`` and ``borderstyle`` will be passed to the scrollbar.
    
    ``content_height`` refers to the maximum offset the user can scroll to.
    
    The content height may be changed, but manually calling :py:meth:`redraw()` will be necessary.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 scrollbar_width=12,
                 borderstyle="flat",
                 content_height=100,
                ):
        self.offset_y = 0
        self.content_height = content_height
        super(ScrollableContainer,self).__init__(name,submenu,window,peng,pos,size,_skip_draw=True)
        self._scrollbar = VerticalSlider("__scrollbar_%s"%name,self,self.peng.window,self.peng,
                                pos=[0,0],
                                size=[24,0],
                                borderstyle=borderstyle,
                                border=[3,3],
                                n=self.content_height,
                                nmax=self.content_height,
                                )
        self._scrollbar._is_scrollbar = True
        self._scrollbar.addAction("progresschange",self.redraw)
        self.addWidget(self._scrollbar)
        
        self.redraw()
    
    def on_redraw(self):
        """
        Redraws the background and contents, including scrollbar.
        
        This method will also check the scrollbar for any movement and will be automatically called on movement of the slider.
        """
        n = self._scrollbar.n
        self.offset_y = -n # Causes the content to move in the opposite direction of the slider
        
        # Size of scrollbar
        sx=24 # Currently constant, TODO: add dynamic sx of scrollbar
        sy=self.size[1]
        # Pos of scrollbar
        x=self.size[0]-sx
        y=0 # Currently constant, TODO: add dynamic y-pos of scrollbar
        
        # Dynamic pos/size may be added via align/lambda/etc.
        
        # Note that the values are written to the _* variant of the attribute to avoid 3 uneccessary redraws
        self._scrollbar._size = sx,sy
        self._scrollbar._pos = x,y
        self._scrollbar._nmax = self.content_height
        
        super(ScrollableContainer,self).on_redraw() # Re-draws everything, including child widgets


# Hack to allow BasicWidget to do isinstance of Container/ScrollableContainer for offset code
from ..gui import widgets
widgets.Container = Container
widgets.ScrollableContainer = ScrollableContainer



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  widgets.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "BasicWidget","Background",
    "Widget","EmptyBackground",
    ]

import weakref
import time

# Internal Debug/Performance monitor variable
_num_saved_redraws = 0

try:
    import pyglet
    from pyglet.gl import *
except ImportError:
    pass # Headless mode

from ..util import mouse_aabb, ActionDispatcher, WatchingList as _WatchingList

class Background(object):
    """
    Class representing the background of a widget.
    
    This base class does not do anything.
    """
    def __init__(self,widget):
        self.widget = widget
        self.initialized = False
        self._vlists = []
    def init_bg(self):
        """
        Called just before the background will be drawn the first time.
        
        Commonly used to initialize vertex lists.
        
        It is recommended to add all vertex lists to the ``submenu.batch2d`` Batch to speed up rendering and preventing glitches with grouping.
        """
        pass
    def redraw_bg(self):
        """
        Method called by the parent widget every time its :py:meth:`Widget.redraw()` method is called.
        """
        pass
    
    def reg_vlist(self,vlist):
        """
        Registers a vertex list to the internal list.
        
        This allows the class to clean itself up properly during deletion, as the background would still be visible after deletion otherwise.
        """
        self._vlists.append(vlist)
    
    @property
    def submenu(self):
        """
        Property for accessing the parent widget's submenu.
        """
        return self.widget.submenu
    
    @property
    def window(self):
        """
        Property for accessing the parent widget's window.
        """
        return self.widget.window
    
    @property
    def peng(self):
        """
        Property for accessing the parent widget's instance of :py:class:`peng3d.peng.Peng`\ .
        """
        return self.widget.peng
    
    @property
    def pressed(self):
        """
        Read-only helper property for easier access.
        
        Equivalent to ``widget.pressed``\ .
        """
        return self.widget.pressed
    @property
    def is_hovering(self):
        """
        Read-only helper property for easier access.
        
        Equivalent to ``widget.is_hovering``\ .
        """
        return self.widget.is_hovering
    
    def __del__(self):
        for vlist in self._vlists:
            try:
                vlist.delete()
            except Exception:
                pass
        self._vlists=[]

class EmptyBackground(Background):
    """
    Background that draws simply nothing.
    
    Can be used as a placeholder.
    """
    pass

class BasicWidget(ActionDispatcher):
    """
    Basic Widget class.
    
    Every widget must be registered with their appropriate sub-menus to work properly.
    
    ``pos`` may be either a list or 2-tuple of ``(x,y)`` for static positions or a function with the signature ``window_width,window_height,widget_width,widget_height`` returning a tuple.
    
    ``size`` is similar to ``pos`` but will only get ``window_width,window_height`` as its arguments.
    
    Commonly, the lambda ``lambda sw,sh,bw,bh: (sw/2.-bw/2.,sh/2.-bh/2.)`` is used to center the widget.
    
    The actions available may differ from widget to widget, by default these are used:
    
    - ``press`` is called upon starting to click on the widget
    - ``click`` is called if the mouse is released on the widget while also having been pressed on it before, recommended for typical button callbacks
    - ``context`` is called upon right-clicking on the widget and may be used to display a context menu
    - ``hover_start`` signals that the cursor is now hovering over the widget
    - ``hover`` is called every time the cursor moves while still being over the widget
    - ``hover_end`` is called after the cursor leaves the widget

    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None):
        self.name = name
        self.submenu = submenu
        self.window = window
        self.peng = peng
        
        self.actions = {}
        
        self._pos = pos
        self._size = size
        
        self.is_hovering = False
        self.pressed = False
        self._enabled = True
        self._is_scrollbar = False
        self.do_redraw = True
        
        self.registerEventHandlers()
    
    def registerEventHandlers(self):
        """
        Registers event handlers used by this widget, e.g. mouse click/motion and window resize.
        
        This will allow the widget to redraw itself upon resizing of the window in case the position needs to be adjusted.
        """
        self.peng.registerEventHandler("on_mouse_press",self.on_mouse_press)
        self.peng.registerEventHandler("on_mouse_release",self.on_mouse_release)
        self.peng.registerEventHandler("on_mouse_drag",self.on_mouse_drag)
        self.peng.registerEventHandler("on_mouse_motion",self.on_mouse_motion)
        self.peng.registerEventHandler("on_resize",self.on_resize)
    
    @property
    def pos(self):
        """
        Property that will always be a 2-tuple representing the position of the widget.
        
        Note that this method may call the method given as ``pos`` in the initializer.
        
        The returned object will actually be an instance of a helper class to allow for setting only the x/y coordinate.
        
        This property also respects any :py:class:`Container` set as its parent, any offset will be added automatically.
        
        Note that setting this property will override any callable set permanently.
        """
        if isinstance(self._pos,list) or isinstance(self._pos,tuple):
            r = self._pos
        elif callable(self._pos):
            w,h = self.submenu.size[:]
            r = self._pos(w,h,*self.size)
        else:
            raise TypeError("Invalid position type")
        
        ox,oy = self.submenu.pos
        r = r[0]+ox,r[1]+oy
        
        #if isinstance(self.submenu,ScrollableContainer) and not self._is_scrollbar:# and self.name != "__scrollbar_%s"%self.submenu.name: # Widget inside scrollable container and not the scrollbar
        #    r = r[0],r[1]+self.submenu.offset_y
        return _WatchingList(r,self._wlredraw_pos)
    @pos.setter
    def pos(self,value):
        self._pos = value
        self.redraw()
    
    @property
    def size(self):
        """
        Similar to :py:attr:`pos` but for the size instead.
        """
        if isinstance(self._size,list) or isinstance(self._size,tuple):
            s = self._size
        elif callable(self._size):
            w,h = self.submenu.size[:]
            s = self._size(w,h)
        else:
            raise TypeError("Invalid size type")
        
        s = s[:]
        
        if s[0]==-1:
            s[0]=self.getMinSize()[0]
        if s[1]==-1:
            s[1]=self.getMinSize()[1]
    
        # Prevents crashes with negative size
        s = [max(s[0],0),max(s[1],0)]
        return _WatchingList(s,self._wlredraw_size)
    @size.setter
    def size(self,value):
        self._size = value
        self.redraw()
    
    @property
    def clickable(self):
        """
        Property used for determining if the widget should be clickable by the user.
        
        This is only true if the submenu of this widget is active and this widget is enabled.
        
        The widget may be either disabled by setting this property or the :py:attr:`enabled` attribute.
        """
        if not isinstance(self.submenu,Container):
            return self.submenu.name == self.submenu.menu.activeSubMenu and self.submenu.menu.name == self.window.activeMenu and self.enabled
        else:
            return self.submenu.clickable and self.enabled
    @clickable.setter
    def clickable(self,value):
        self._enabled=value
        self.redraw()
    
    @property
    def enabled(self):
        """
        Property used for storing whether or not this widget is enabled.
        
        May influence rendering and behavior.
        
        Note that the widget will be immediately redrawn if this property is changed.
        """
        return self._enabled
    @enabled.setter
    def enabled(self,value):
        self._enabled=value
        self.redraw()
    
    def getState(self):
        """
        Returns the current state of the widget.
        
        One of ``"pressed"``\ , ``"hover"``\ , ``"disabled"`` or ``"idle"``\ .
        Note that some information may be lost by getting this state, 
        for example it is not possible to know if the widget is hovered or not
        if ``"pressed"`` is returned. However, this should not be a problem for
        most intended uses of this method.
        """
        # Does not convey all information, loses some
        if self.pressed:
            return "pressed"
        elif self.is_hovering:
            return "hover"
        elif not self.enabled:
            return "disabled"
        else:
            return "idle"
    
    def getMinSize(self):
        raise NotImplementedError("%s does not support perfect-size widgets")
    def draw(self):
        """
        Draws all vertex lists associated with this widget.
        """
        if self.do_redraw:
            self.on_redraw()
            self.do_redraw = False
    def redraw(self):
        """
        Triggers a redraw of the widget.
        
        Note that the redraw may not be executed instantly, but rather batched together on the next frame.
        If an instant and synchronous redraw is needed, use :py:meth:`on_redraw()` instead.
        """
        # Uncomment if you want to see how many unneccessary redraws have been prevented
        # Warning: may cause lots of console spam and thus inaccurate results
        #if self.do_redraw:
        #    # Debug - not for production use
        #    global _num_saved_redraws
        #    _num_saved_redraws+=1
        #    print("saved redraw #%s"%_num_saved_redraws)
        self.do_redraw = True
    def on_redraw(self):
        """
        Callback to be overridden by subclasses called if redrawing the widget seems necessary.
        
        Note that this method should not be called manually, see :py:meth:`redraw()` instead.
        """
        pass
    
    def _wlredraw_pos(self,wl):
        self._pos = wl[:]
        self.redraw()
    def _wlredraw_size(self,wl):
        self._size = wl[:]
        self.redraw()
    
    def on_mouse_press(self,x,y,button,modifiers):
        if not self.clickable:
            return
        elif mouse_aabb([x,y],self.size,self.pos) and mouse_aabb([x,y],self.submenu.size,self.submenu.pos):
            if button == pyglet.window.mouse.LEFT:
                self.doAction("press")
                self.pressed = True
                self.doAction("statechanged")
            elif button == pyglet.window.mouse.RIGHT:
                self.doAction("context")
            self.redraw()
    def on_mouse_release(self,x,y,button,modifiers):
        if not self.clickable:
            return
        if self.pressed:
            if mouse_aabb([x,y],self.size,self.pos):
                self.doAction("click")
            self.pressed = False
            self.doAction("statechanged")
            self.redraw()
    def on_mouse_drag(self,x,y,dx,dy,button,modifiers):
        self.on_mouse_motion(x,y,dx,dy)
    def on_mouse_motion(self,x,y,dx,dy):
        if not self.clickable:
            return
        if mouse_aabb([x,y],self.size,self.pos):
            if not self.is_hovering:
                self.is_hovering = True
                self.doAction("hover_start")
                self.doAction("statechanged")
                self.redraw()
            else:
                self.doAction("hover")
        else:
            if self.is_hovering:
                self.is_hovering = False
                self.doAction("hover_end")
                self.doAction("statechanged")
                self.redraw()
    def on_resize(self,width,height):
        self.redraw()
    
    def delete(self):
        """
        Deletes resources of this widget that require manual cleanup.
        
        Currently removes all actions, event handlers and the background.
        
        The background itself should automatically remove all vertex lists to avoid visual artifacts.
        
        Note that this method is currently experimental, as it seems to have a memory leak.
        """
        # TODO: fix memory leak upon widget deletion
        del self.bg.widget
        del self.bg
        
        #self.clickable=False
        
        del self._pos
        del self._size
        
        self.actions = {}
        
        for e_type,e_handlers in self.peng.eventHandlers.items():
            if True or e_type in eh:
                to_del = []
                for e_handler in e_handlers:
                    # Weird workaround due to implementation details of WeakMethod
                    if isinstance(e_handler,weakref.ref): 
                        if super(weakref.WeakMethod,e_handler).__call__() is self: 
                            to_del.append(e_handler) 
                    elif e_handler is self: 
                        to_del.append(e_handler)
                for d in to_del:
                    try:
                        #print("Deleting handler %s of type %s"%(d,e_type))
                        del e_handlers[e_handlers.index(d)]
                    except Exception:
                        #print("Could not delete handler %s, memory leak may occur"%d)
                        import traceback;traceback.print_exc()
    #def __del__(self):
    #    print("del %s"%self.name)
    #    try:
    #        del self.bg
    #    except Exception:
    #        pass

class Widget(BasicWidget):
    """
    Subclass of :py:class:`BasicWidget` adding support for changing the :py:class:`Background`\ .
    
    If no background is given, an :py:class:`EmptyBackground` will be used instead.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 bg=None,
                 min_size=None,
                ):
        if bg is None:
            bg = EmptyBackground(self)
        self.bg = bg
        self.min_size = min_size
        super(Widget,self).__init__(name,submenu,window,peng,pos,size)
    def setBackground(self,bg):
        """
        Sets the background of the widget.
        
        This may cause the background to be initialized.
        """
        self.bg = bg
        self.redraw()
    
    def getMinSize(self):
        c = self.getContentSize()
        m = self.min_size if self.min_size is not None else [0,0]
        return [max(c[0],m[0]),max(c[1],m[1])]
    def getContentSize(self):
        return [0,0]
    
    def on_redraw(self):
        """
        Draws the background and the widget itself.
        
        Subclasses should use ``super()`` to call this method, or rendering may glitch out.
        """
        if self.bg is not None:
            if not self.bg.initialized:
                self.bg.init_bg()
                self.bg.initialized=True
            self.bg.redraw_bg()
        super(Widget,self).on_redraw()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  version.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["VERSION","RELEASE"]

VERSION = "1.10.1"
"""
Full version number of format ``MAJOR.MINOR.BUGFIX`` where major is increased only on very major feature changes.
Minor is changed if a new feature is introduced or an API change is made, while bugfix only changes if an important fix for a bug needs to be provided before the next release.

Used to display the version in the title of the documentation.

.. seealso::
   
   See the `Distutils docs on version numbers <https://docs.python.org/2/distutils/setupscript.html#additional-meta-data>`_ for more information.

"""

RELEASE = "1.10.1"
"""
Currently the same as :py:data:`VERSION`\ .

Used to display the version on the top-right of the documentation.
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  model.py
#  
#  Copyright 2017 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "grouper","calcSphereCoordinates",
    "v_magnitude","v_normalize",
    "Material", "Bone", "RootBone",
    "Region", "Animation",
    "JSONModelGroup","JSONRegionGroup",
    "Model",
    ]

import time

import math
from math import sqrt,cos,sin

try:
    from itertools import zip_longest
except ImportError:
    pass # probably on readthedocs

try:
    import pyglet
    from pyglet.gl import *
except ImportError:
    pass # probably headless

def grouper(iterable, n, fillvalue=None):
    """
    Allows for iteration over multiple elements of a iterable at once.
    
    ``iterable`` may be any iterable, its values will be returned. Note that this may be iterated over more than once.
    
    ``n`` is the size of each group. May be any positive integer.
    
    ``fillvalue`` is optionally used to fill any groups that do not have enough items, for example if the length of the iterable is not divisible by n.
    
    Example::
        
        >>> for i in grouper("foobarbaz",2,fillvalue=" "):
        ...     print(i)
        fo
        ob
        ar
        ba
        z  # Note the extra space after the z
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)

def calcSphereCoordinates(pos,radius,rot):
    """
    Calculates the Cartesian coordinates from spherical coordinates.
    
    ``pos`` is a simple offset to offset the result with.
    
    ``radius`` is the radius of the input.
    
    ``rot`` is a 2-tuple of ``(azimuth,polar)`` angles.
    
    Angles are given in degrees. Most directions in this game use the same convention.
    
    The azimuth ranges from 0 to 360 degrees with 0 degrees pointing directly to the x-axis.
    
    The polar angle ranges from -90 to 90 with -90 degrees pointing straight down and 90 degrees straight up.
    
    A visualization of the angles required is given in the source code of this function.
    """
    # Input angles should be in degrees, as in the rest of the game
    # E.g. phi=inclination and theta=azimuth
    
    # phi is yrad
    #      Look from above         
    #(Z goes positive towards you) 
    #                              
    #            Y-  Z-            
    #            |  /              
    #            | / "far"         
    #            |/                
    #   X- ------+-------> X+      
    #           /| yrad |          
    #   "near" / |<-----+          
    #         /  |  "polar angle"  
    #        Z+  Y+                
    
    # theta is xrad
    #      Look from above         
    #(Z goes positive towards you) 
    #                              
    #            Y-  Z-            
    #            |  /              
    #            | / "far"         
    #            |/                
    #   X- ------+-------> X+      
    #           /| xrad |          
    #   "near" /<-------+          
    #         /  |  "azimuth angle"
    #        Z+  Y+                

    # Based on http://stackoverflow.com/questions/39647735/calculation-of-spherical-coordinates
    # https://en.wikipedia.org/wiki/Spherical_coordinate_system
    # http://stackoverflow.com/questions/25404613/converting-spherical-coordinates-to-cartesian?rq=1
    
    phi,theta = rot
    phi+=90 # very important, took me four days of head-scratching to figure out
    phi,theta = math.radians(phi),math.radians(theta)
    
    x = pos[0]+radius * math.sin(phi) * math.cos(theta)
    y = pos[1]+radius * math.sin(phi) * math.sin(theta)
    z = pos[2]+radius * math.cos(phi)
    return x,y,z

def v_magnitude(v):
    """
    Simple vector helper function returning the length of a vector.
    
    ``v`` may be any vector, with any number of dimensions
    """
    return math.sqrt(sum(v[i]*v[i] for i in range(len(v))))

def v_normalize(v):
    """
    Normalizes the given vector.
    
    The vector given may have any number of dimensions.
    """
    vmag = v_magnitude(v)
    return [ v[i]/vmag  for i in range(len(v)) ]

class Material(object):
    """
    Object that describes a single material of a model.
    
    This object stores all relevant data and caches. Note that this object is only created once for each model and shared between all rendered instances of it.
    
    See :py:class:`Model` for more information about the model system.
    """
    def __init__(self,rsrcMgr,name,matdata):
        self.rsrcMgr = rsrcMgr
        
        self.name = name
        self.matdata = matdata
        
        self.texname = matdata.get("tex","peng3d:missingtexture")
        self.texcat = matdata.get("texcat","entity")
        if self.texcat not in self.rsrcMgr.categories:
            self.rsrcMgr.addCategory(self.texcat)
        
    @property
    def target(self):
        """
        Read-only property storing the OpenGL constant representing the target of the texture of this material.
        
        Commonly :py:const:`GL_TEXTURE_2D` or :py:const:`GL_TEXTURE_3D`\ .
        
        Used in texture manipulation and activation, e.g. ``glEnable(material.target)``\ .
        """
        return self.rsrcMgr.getTex(self.texname,self.texcat)[0]
    @property
    def id(self):
        """
        Read-only property storing the numerical ID of the texture of this material.
        
        Used to manipulate the texture behind this material.
        
        Commonly used in binding the texture: ``glBindTexture(material.target,material.id)``\ .
        """
        return self.rsrcMgr.getTex(self.texname,self.texcat)[1]
    @property
    def tex_coords(self):
        """
        Read-only property storing the texture coordinates to use when drawing with this texture.
        
        Should not be used directly, see :py:meth:`transformTexCoords()`\ .
        
        Enables substitution of pyglet :py:class:`pyglet.graphics.Texture` objects with Materials in many places, e.g. in :py:class:`pyglet.graphics.TextureGroup`\ .
        """
        return self.rsrcMgr.getTex(self.texname,self.texcat)[2]
    
    @property
    def texdata(self):
        """
        Read-only property equivalent to a 3-tuple containing :py:attr:`target`\ , :py:attr:`id` and :py:attr:`tex_coords`\ .
        
        Should be faster than getting each value directly. Useful if all of these values are needed.
        """
        return self.rsrcMgr.getTex(self,texname,self.texcat)
    
    def transformTexCoords(self,data,texcoords,dims=2):
        """
        Transforms the given texture coordinates using the internal texture coordinates.
        
        Currently, the dimensionality of the input texture coordinates must always be 2 and the output is 3-dimensional with the last coordinate always being zero.
        
        The given texture coordinates are fitted to the internal texture coordinates. Note that values higher than 1 or lower than 0 may result in unexpected visual glitches.
        
        The length of the given texture coordinates should be divisible by the dimensionality.
        """
        assert dims==2 # TODO
        
        out = []
        
        origcoords = self.tex_coords
        min_u,min_v = origcoords[0],origcoords[1]
        max_u,max_v = origcoords[6],origcoords[7]
        
        diff_u,diff_v = max_u-min_u, max_v-min_v

        itexcoords = iter(texcoords)
        for u,v in zip(itexcoords,itexcoords): # Based on http://stackoverflow.com/a/5389547/3490549
            out_u = min_u+(diff_u*u)
            out_v = min_v+(diff_v*v)
            out.extend((out_u,out_v,0))
        
        return out # output dimensionality is 3, e.g. t3f

class Bone(object):
    """
    Object that represents a single bone of a model.
    
    This object stores all relevant data and caches. Note that this object is only created once for each model and shared between all rendered instances of it.
    
    Actual bone rotation and length is stored per entity and not per model allowing for different bone rotations for multiple entities using the same model.
    
    See :py:class:`Model` for more information about the model system.
    """
    def __init__(self,rsrcMgr,name,bonedata):
        self.rsrcMgr = rsrcMgr
        
        self.name = name
        self.bonedata = bonedata
        
        self.regions = {}
        self.child_bones = {}
        
        self.parent = None
        self.start_rot = bonedata.get("start_rot",[0,0])
        self.blength = bonedata.get("length",1.0)
    
    def ensureBones(self,data):
        """
        Helper method ensuring per-entity bone data has been properly initialized.
        
        Should be called at the start of every method accessing per-entity data.
        
        ``data`` is the entity to check in dictionary form.
        """
        if "_bones" not in data:
            data["_bones"]={}
        if self.name not in data["_bones"]:
            data["_bones"][self.name]={"rot":self.start_rot[:],"length":self.blength}
    
    def setRot(self,data,rot):
        """
        Sets the rotation of this bone on the given entity.
        
        ``data`` is the entity to modify in dictionary form.
        
        ``rot`` is the rotation of the bone in the format used in :py:func:`calcSphereCoordinates()`\ .
        """
        self.ensureBones(data)
        rot = rot[0]%360,max(-90, min(90, rot[1]))
        data["_bones"][self.name]["rot"]=rot
    def getRot(self,data):
        """
        Returns the rotation of this bone in the given entity.
        
        ``data`` is the entity to query in dictionary form.
        """
        self.ensureBones(data)
        return data["_bones"][self.name]["rot"]
    
    def setLength(self,data,blength):
        """
        Sets the length of this bone on the given entity.
        
        ``data`` is the entity to modify in dictionary form.
        
        ``blength`` is the new length of the bone.
        """
        self.ensureBones(data)
        data["_bones"][self.name]["length"]=blength
    def getLength(self,data):
        """
        Returns the length of this bone in the given entity.
        
        ``data`` is the entity to query in dictionary form.
        """
        self.ensureBones(data)
        return data["_bones"][self.name]["length"]
    
    # No properties to access rot/length directly as they are entity-dependent
    
    def setParent(self,parent):
        """
        Sets the parent of this bone for all entities.
        
        Note that this method must be called before many other methods to ensure internal state has been initialized.
        
        This method also registers this bone as a child of its parent.
        """
        self.parent = parent
        self.parent.child_bones[self.name]=self
    
    def setRotate(self,data):
        """
        Sets the OpenGL state required for proper drawing of the model.
        
        Mostly rotates and translates the camera.
        
        It is important to call :py:meth:`unsetRotate()` after calling this method to properly unset state and avoid OpenGL errors.
        """
        self.parent.setRotate(data)
        glPushMatrix()
        x,y = self.getRot(data)
        pivot = self.getPivotPoint(data)
        px,py,pz = pivot

        #ppx,ppy,ppz = self.parent.getPivotPoint(data)
        #rot_vec = px-ppx,py-ppy,pz-ppz
        #print([px,py,pz],[ppx,ppy,ppz],rot_vec,[x,y],self.getLength(data))
        #norm = v_normalize(rot_vec,self.name) if v_magnitude(rot_vec)==0 else [0,1,0] # prevents a zero divison error
        normx = [0,1,0] # currently fixed to the up vector, should be changed in the future for proper rotation that is not along the y or z axis
        normy = [0,0,1]
        
        #offset = -px,-py,-pz
        
        glTranslatef(px,py,pz)#-offset[0],-offset[1],-offset[2])
        glRotatef(x-self.start_rot[0], normx[0],normx[1],normx[2])
        glRotatef(y-self.start_rot[1], normy[0],normy[1],normy[2])
        glTranslatef(-px,-py,-pz)#offset[0],offset[1],offset[2])
    def unsetRotate(self,data):
        """
        Unsets the OpenGL state that was set before calling :py:meth:`setRotate()`\ .
        
        Note that this method may cause various OpenGL errors if called without :py:meth:`setRotate()` having been called.
        """
        glPopMatrix()
        self.parent.unsetRotate(data)
    
    def getPivotPoint(self,data):
        """
        Returns the point this bone pivots around on the given entity.
        
        This method works recursively by calling its parent and then adding its own offset.
        
        The resulting coordinate is relative to the entity, not the world.
        """
        ppos = self.parent.getPivotPoint(data)
        rot = self.parent.getRot(data)
        length = self.parent.getLength(data)
        out = calcSphereCoordinates(ppos,length,rot)
        return out
        
    def transformVertices(self,data,vertices,dims=3):
        """
        Currently unused method that transforms the given vertices according to the rotation of the bone.
        
        Currently just returns the vertices unmodified, will be implemented in the future.
        """
        assert dims==3
        
        # vertices are not transformed directly, rather they are rotated dynamically via glRotate
        
        return vertices 
    
    def addRegion(self,region):
        """
        Register a vertex Region as a dependent of this bone.
        
        ``region`` must be an instance of :py:class:`Region`\ .
        """
        self.regions[region.name]=region

class RootBone(Bone):
    """
    Special bone that represents the root of a entity.
    
    This bone is immutable and cannot be rotated or otherwise modified.
    """
    def getPivotPoint(self,data):
        return [0,0,0]
    getPivotPoint.__noautodoc__ = True
    def getLength(self,data):
        return 0
    getLength.__noautodoc__ = True
    def setRotate(self,data):
        pass # no rotation as the root bone is always pointed straight up
    setRotate.__noautodoc__ = True
    def unsetRotate(self,data):
        pass # no unset as there is no set
    unsetRotate.__noautodoc__ = True

class Region(object):
    """
    Object that represents a vertex region of a model.
    
    A vertex region is associated with a specific bone of the same model it is associated with.
    It has a list of vertices and optionally texture coordinates. The texture coordinates are transformed using the material it is associated with.
    
    Most regions will use quads as their primitive type, but it is also possible to use triangles, lines and points.
    
    To use quads as the geometry type, specify either ``quads``\ , ``quad`` or ``GL_QUADS`` as its ``geometry_type``\ .
    
    To use triangles as the geometry type, specify either ``tris``\ , ``triangles``\ , ``triangle`` or ``GL_TRIANGLES`` as its ``geometry_type``\ .
    
    To use lines as the geometry type, specify either ``lines``\ , ``line`` or ``GL_LINES`` as its ``geometry_type``\ .
    
    To use points as the geometry type, specify either ``points``\ , ``point``\ , ``dots``\, ``dot`` or ``GL_POINTS`` as its ``geometry_type``\ .
    
    Note that the number of vertices must be divisible by the number of vertices required per primitive, e.g. 4 for quads, 3 for triangles, 2 for lines and 1 for points.
    
    Additionally, the number of vertices and texture coordinate pairs must also match.
    
    If any of these conditions are not fulfilled, a :py:exc:`ValueError` will be raised.
    """
    def __init__(self,rsrcMgr,name,regdata):
        self.rsrcMgr = rsrcMgr
        
        self.name = name
        self.regdata = regdata
        
        self.material = None
        self.bone = None
        
        self.dims = 3
        self.tex_dims = 2
        
        gtype = regdata.get("geometry_type","quads")
        if gtype in ["quads","quad","GL_QUADS"]:
            self.geometry_type = GL_QUADS
            # points per primitive
            ppp = 4
        elif gtype in ["tris","triangles","triangle","GL_TRIANGLES"]:
            self.geometry_type = GL_TRIANGLES
            ppp = 3
        elif gtype in ["lines","line","GL_LINES"]:
            self.geometry_type = GL_LINES
            ppp = 2
        elif gtype in ["points","point","dots","dot","GL_POINTS"]:
            self.geometry_type = GL_POINTS
            ppp = 1
        else:
            raise ValueError("Invalid Geometry type %s"%gtype)
        
        self.vertices = regdata.get("vertices",[])
        if len(self.vertices)%self.dims!=0:
            raise ValueError("Vertices must be in x,y,z groups")
        elif (len(self.vertices)/self.dims)%ppp!=0:
            raise ValueError("Invalid amount of x,y,z groups, must be integer-divisible by %s"%ppp)
        
        if "tex_coords" in regdata:
            self.enable_tex = True
            self.tex_coords = regdata.get("tex_coords",[])
            if len(self.tex_coords)/self.tex_dims!=len(self.vertices)/self.dims:
                raise ValueError("Non-Matching amount of vertices and tex coords")
            elif len(self.tex_coords)%self.tex_dims!=0:
                raise ValueError("Tex coords must be in u,v pairs")
            elif (len(self.tex_coords)/self.tex_dims)%ppp!=0:
                raise ValueError("Invalid amount of u,v pairs, must be integer-divisible by %s"%ppp)
        else:
            self.enable_tex = False
            self.tex_coords = [0]*(self.tex_dims*(len(self.vertices)/self.dims))
    
    def getVertices(self,data):
        """
        Returns the vertices of this region already transformed and ready-to-use.
        
        Internally uses :py:meth:`Bone.transformVertices()`\ .
        """
        return self.bone.transformVertices(data,self.vertices,self.dims)
    def getGeometryType(self,data):
        """
        Returns the OpenGL constant representing the type of primitives used by this region.
        
        May be one of :py:data:`GL_QUADS`\ , :py:data:`GL_TRIANGLES`\ , :py:data:`GL_LINES` or :py:data:`GL_POINTS`\ .
        """
        return self.geometry_type
    def getTexCoords(self,data):
        """
        Returns the texture coordinates, if any, to accompany the vertices of this region already transformed.
        
        Note that it is recommended to check the :py:attr:`enable_tex` flag first.
        
        Internally uses :py:meth:`Material.transformTexCoords()`\ .
        """
        return self.material.transformTexCoords(data,self.tex_coords,self.tex_dims)
    def getTexInfo(self,data):
        """
        Returns informations about the texture of this region.
        
        Internally uses :py:attr:`Material.texdata`\ , exact specification available there.
        """
        return self.material.texdata

class Animation(object):
    """
    Object that represents an animation of a model.
    
    Animations can be either static or animated using keyframes.
    
    See :py:class:`Model` for more information.
    """
    def __init__(self,rsrcMgr,name,anidata):
        self.rsrcMgr = rsrcMgr
        
        self.name = name
        self.anidata = anidata
        
        self.bones = {}
        
        self.atype = anidata.get("type","bone" if "keyframes" in anidata else "static")
        
        self.default_jt = anidata.get("default_jumptype","animate")
        
        self.entity_defaults = {}
        self.entity_template = {}
        
        if self.atype == "static":
            self.start_frame = {"bones":anidata.get("bones")}
        elif self.atype == "keyframes":
            self.kps = anidata.get("keyframespersecond",60)
            self.anilength = anidata.get("length",self.kps)
            
            self.lframes_per_bone = {} # for length-setting keyframes
            self.rframes_per_bone = {} # for rotating keyframes
            
            minframe = None
            for frame,keyframe in anidata.get("keyframes",{}).items():
                frame = int(frame)
                if minframe is None or frame<minframe:
                    minframe = frame
                
                for bname,bone in keyframe.get("bones",{}).items():
                    if "rot" in bone:
                        if bname not in self.rframes_per_bone:
                            self.rframes_per_bone[bname]={}
                        self.rframes_per_bone[bname][frame]=bone["rot"]
                    if "length" in bone:
                        if bname not in self.lframes_per_bone:
                            self.lframes_per_bone[bname]={}
                        self.lframes_per_bone[bname][frame]=bone["length"]
            
            self.start_frame = anidata["keyframes"][str(minframe)] if minframe is not None else {"bones":{}}
            self.start_frame["frame"] = minframe if minframe is not None else 0
        else:
            raise ValueError("Invalid animation type %s for animation %s"%(self.atype,name))
    
    def setBones(self,bones):
        """
        Sets the internal dictionary of bones in the parent model.
        
        Must be a dictionary, else errors may appear later on.
        """
        self.bones = bones
    
    def startAnimation(self,data,jumptype):
        """
        Callback that is called to initialize this animation on a specific actor.
        
        Internally sets the ``_anidata`` key of the given dict ``data``\ .
        
        ``jumptype`` is either ``jump`` or ``animate`` to define how to switch to this animation.
        """
        data["_anidata"]={}
        adata = data["_anidata"]
        
        adata["keyframe"]=0
        adata["last_tick"]=time.time()
        adata["jumptype"]=jumptype
        adata["phase"]="transition"
    
    def tickEntity(self,data):
        """
        Callback that should be called regularly to update the animation.
        
        It is recommended to call this method about 60 times a second for smooth animations. Irregular calling of this method will be automatically adjusted.
        
        This method sets all the bones in the given actor to the next state of the animation.
        
        Note that :py:meth:`startAnimation()` must have been called before calling this method.
        """
        adata = data["_anidata"]
        
        if adata.get("anitype",self.name)!=self.name:
            return # incorrectly called
        
        dt = time.time()-adata.get("last_tick",time.time())
        adata["last_tick"]=time.time()
        
        if adata.get("phase","transition")=="transition":
            # If transitioning to this animation
            if adata.get("jumptype",self.default_jt)=="jump":
                # Jumping to the animation, e.g. set all bones to the start frame
                for bone,dat in self.start_frame.get("bones",{}).items():
                    if "rot" in dat:
                        self.bones[bone].setRot(data,dat["rot"])
                    if "length" in dat:
                        self.bones[bone].setLength(data,dat["length"])
                adata["phase"]="animation"
                if self.atype=="keyframes":
                    adata["from_frame"]=self.start_frame.get("frame",0)
            elif adata.get("jumptype",self.default_jt)=="animate":
                raise NotImplementedError("Animation update phase transition type animate not yet implemented")
                
                # Not yet working
                if "transition_frame" not in adata:
                    adata["transition_frame"] = 0
                tlen = self.anidata.get("transition_length",self.anidata.get("length",60)/3)
                tspeed = self.anidata.get("transition_speed",self.anidata.get("keyframespersecond",60))
                
                framediff=int(dt/(1./tspeed)) # Number of frames that have passed since the last update
                overhang = dt-(framediff*(1./tspeed)) # time that has passed but is not enough for a full frame
                adata["last_tick"]-=overhang # causes the next frame to include the overhang from this frame
                if framediff == 0:
                    return # optimization that saves time 
                
                frame1 = adata["transition_frame"]
                adata["transition_frame"]+=framediff
                if adata["transition_frame"]>tlen:
                    adata["phase"]="animation"
                    if self.atype=="keyframes":
                        adata["from_frame"]=self.start_frame.get("frame",0)
                    return
                frame2 = adata["transition_frame"]
                
                if "frombones" not in adata:
                    frombones = {}
                    for bname,bone in self.bones.items():
                        frombones[bname]={"rot":bone.getRot(data),"length":bone.getLength(data)}
                    adata["frombones"] = frombones
                if "tobones" not in adata:
                    tobones = {}
                    for bname,bone in self.start_frame["bones"].items():
                        tobones[bname]=bone
                    adata["tobones"] = tobones
                
                frombones = adata["frombones"]
                tobones = adata["tobones"]
                
                from_frame = 0
                to_frame = tlen
                
                for bname,bone in self.bones.items():
                    # rot
                    if bname not in frombones or bname not in tobones:
                        continue
                    
                    from_rot = frombones[bname]["rot"]
                    to_rot = tobones[bname]["rot"]
                    
                    from_x,from_y=from_rot
                    to_x,to_y=to_rot
                    
                    delta_per_frame_x = (to_x-from_x)/(to_frame-from_frame)
                    delta_per_frame_y = (to_y-from_y)/(to_frame-from_frame)
                    
                    delta_x = delta_per_frame_x*(frame2-frame1)
                    delta_y = delta_per_frame_y*(frame2-frame1)
                    
                    rot = bone.getRot(data)
                    new_rot = [rot[0]+delta_x,rot[1]+delta_y]
                    
                    bone.setRot(data,new_rot)
                    
                    # length
                    
                    # Not yet implemented
                
        elif adata["phase"]=="animation":
            # If animating
            if self.atype == "static":
                pass # Should not need any updates as the transition will set the bones
            elif self.atype == "keyframes":
                framediff=int(dt/(1./self.kps)) # Number of frames that have passed since the last update
                overhang = dt-(framediff*(1./self.kps)) # time that has passed but is not enough for a full frame
                adata["last_tick"]-=overhang # causes the next frame to include the overhang from this frame
                if framediff == 0:
                    return # optimization that saves time 
                
                frame1 = adata["keyframe"]
                adata["keyframe"]+=framediff
                if adata["keyframe"]>self.anilength:
                    repeat = True
                    adata["keyframe"]%=self.anilength
                else:
                    repeat = False
                frame2 = adata["keyframe"]
                
                if repeat:
                    frame1 = frame2
                    frame2+=1
                
                for bname,bone in self.bones.items():
                    # Rot
                    if bname in self.rframes_per_bone:
                        
                        from_frame = None#adata["from_frame"]
                        to_frame = None
                        for framenum,rot in self.rframes_per_bone[bname].items():
                            if (from_frame is None or framenum>from_frame) and framenum<=frame1:
                                # from_frame is the largest frame number that is before the starting point
                                from_frame = framenum
                            if (to_frame is None or framenum<to_frame) and framenum>=frame2:
                                # to_frame is the smallest frame number that is after the end point
                                to_frame = framenum
                        if from_frame is None or to_frame is None:
                            raise ValueError("Invalid frames for bone %s in animation %s"%(bname,self.name))
                        if from_frame==to_frame or repeat:
                            if self.anidata.get("repeat","jump")=="jump":
                                for b,dat in self.start_frame.get("bones",{}).items():
                                    if "rot" in dat:
                                        self.bones[b].setRot(data,dat["rot"])
                                    if "length" in dat:
                                        self.bones[b].setLength(data,dat["length"])
                            elif self.anidata.get("repeat","jump")=="animate":
                                from_frame = adata["to_frame"]
                        adata["from_frame"]=from_frame
                        adata["to_frame"]=to_frame
                        
                        from_rot = self.rframes_per_bone[bname][from_frame]
                        to_rot = self.rframes_per_bone[bname][to_frame]
                        
                        from_x,from_y=from_rot
                        to_x,to_y=to_rot
                        
                        if self.anidata.get("interpolation","linear") == "linear":
                            delta_per_frame_x = (to_x-from_x)/(to_frame-from_frame)
                            delta_per_frame_y = (to_y-from_y)/(to_frame-from_frame)
                            
                            delta_x = delta_per_frame_x*(frame2-frame1)
                            delta_y = delta_per_frame_y*(frame2-frame1)
                            
                            rot = bone.getRot(data)
                            new_rot = [rot[0]+delta_x,rot[1]+delta_y]
                        elif self.anidata["interpolation"] == "jump":
                            new_rot = to_x,to_y
                        else:
                            raise ValueError("Invalid interpolation method '%s' for animation %s"%(self.anidata["interpolation"],self.name))
                        
                        bone.setRot(data,new_rot)
                        
                    # Length
                    if bname in self.lframes_per_bone:
                        # Not yet implemented
                        pass
                
            else:
                # Should not be possible, but still
                raise ValueError("Invalid animation type %s for animation %s during tick"%(self.atype,name))


class JSONModelGroup(pyglet.graphics.Group):
    """
    Pyglet group that sets the state required by a specific actor.
    
    This group should always be set during any draw operations for the assigned actor.
    This can either be done by setting it as the group of a vertex list,
    the parent group of a group of a vertex list
    or manually calling :py:meth:`set_state()` and :py:meth:`unset_state()`\ .
    """
    def __init__(self,model,data,obj,parent=None):
        super(JSONModelGroup,self).__init__(parent)
        
        self.model = model
        self.data = data
        self.obj = obj
    
    def set_state(self):
        """
        Sets the state required for this actor.
        
        Currently translates the matrix to the position of the actor.
        """
        x,y,z = self.obj.pos
        glTranslatef(x,y,z)
    
    def unset_state(self):
        """
        Resets the state required for this actor to the default state.
        
        Currently resets the matrix to its previous translation.
        """
        x,y,z = self.obj.pos
        glTranslatef(-x,-y,-z)
    
    def __hash__(self):
        # Internal method to allow unification of multiple groups of the same actor
        return hash((self.obj,))
    
    def __eq__(self,other):
        return (self.__class__ is other.__class__ and
                self.obj.pos == other.obj.pos and
                self.parent == other.parent
               )

class JSONRegionGroup(pyglet.graphics.Group):
    """
    Pyglet group that manages the state required by a specific vertex region of an actor.
    
    This group and the associated :py:class:`JSONModelGroup` should always be set during any draw operation for the assigned region.
    
    See :py:class:`JSONModelGroup` for more information about how to do this.
    """
    def __init__(self,model,data,region,parent=None):
        super(JSONRegionGroup,self).__init__(parent)
        
        self.model = model
        self.data = data
        self.region = region
    
    def set_state(self):
        """
        Sets the state required for this vertex region.
        
        Currently binds and enables the texture of the material of the region.
        """
        glEnable(self.region.material.target)
        glBindTexture(self.region.material.target, self.region.material.id)
        self.region.bone.setRotate(self.data)
    
    def unset_state(self):
        """
        Resets the state required for this actor to the default state.
        
        Currently only disables the target of the texture of the material, it may still be bound.
        """
        glDisable(self.region.material.target)
        self.region.bone.unsetRotate(self.data)
    
    def __hash__(self):
        return hash((self.region.material.target, self.region.material.id, self.parent))

    def __eq__(self, other):
        return (self.__class__ is other.__class__ and
            self.region.material.target == other.region.material.target and
            self.region.material.id == other.region.material.id and
            self.region.bone is other.region.bone and
            self.region.bone.getRot(self.data) == other.region.bone.getRot(other.data) and
            self.region.bone.getLength(self.data) == other.region.bone.getLength(other.data) and # TODO: include parent bones
            self.parent == other.parent
            )


class Model(object):
    """
    Object that represents the model of an actor.
    
    Note that this object is not bound to an actor but rather to a collection of materials, bones, vertex regions and animations.
    
    A single instance of this class may be used by multiple actors at the same time. See :py:meth:`Actor.setModel()` for more information.
    
    A test model is available at ``assets/peng3d/model/test.json`` and a demo program using it under ``test_model.py``\ .
    
    .. todo::
       
       Document the format of .json model files.
    """
    def __init__(self,peng,rsrcMgr,name):
        self.peng = peng
        self.rsrcMgr = rsrcMgr
        self.name = name
        
        self.modeldata = self.rsrcMgr.getModelData(name)
    
    def ensureModelData(self,obj):
        """
        Ensures that the given ``obj`` has been initialized to be used with this model.
        
        If the object is found to not be initialized, it will be initialized.
        """
        if not hasattr(obj,"_modeldata"):
            self.create(obj,cache=True)
        if "_modelcache" not in obj._modeldata:
            # Assume all initialization is missing, simply reinitialize
            self.create(obj,cache=True)
    
    def create(self,obj,cache=False):
        """
        Initializes per-actor data on the given object for this model.
        
        If ``cache`` is set to True, the entity will not be redrawn after initialization.
        
        Note that this method may set several attributes on the given object, most of them starting with underscores.
        
        During initialization of vertex regions, several vertex lists will be created.
        If the given object has an attribute called ``batch3d`` it will be used, else it will be created.
        
        If the batch already existed, the :py:meth:`draw()` method will do nothing, else it will draw the batch.
        
        Memory leaks may occur if this is called more than once on the same object without calling :py:meth:`cleanup()` first.
        """
        obj._modeldata = {}
        
        data = obj._modeldata
        
        data["_model"]=self
        data["_modelcache"] = {}
        moddata = data["_modelcache"]
        
        # Model group, for pos of entity
        moddata["group"] = JSONModelGroup(self,data,obj)
        modgroup = moddata["group"]
        
        if not hasattr(obj,"batch3d"):
            obj.batch3d = pyglet.graphics.Batch()
            data["_manual_render"]=True
        
        # Vlists/Regions
        moddata["vlists"] = {}
        for name,region in self.modeldata["regions"].items():
            v = region.getVertices(data)
            vlistlen = int(len(v)/region.dims)
            
            if region.enable_tex:
                vlist = obj.batch3d.add(vlistlen,region.getGeometryType(data),JSONRegionGroup(self,data,region,modgroup),
                        "v3f/static",
                        "t3f/static",
                        )
            else:
                vlist = obj.batch3d.add(vlistlen,region.getGeometryType(data),modgroup,
                        "v3f/static",
                        )
            
            moddata["vlists"][name]=vlist
        
        self.setAnimation(obj,self.modeldata["default_animation"].name,transition="jump")
        
        self.data = data
        
        if not cache:
            self.redraw(obj)
    
    def cleanup(self,obj):
        """
        Cleans up any left over data structures, including vertex lists that reside in GPU memory.
        
        Behaviour is undefined if it is attempted to use this model with the same object without calling :py:meth:`create()` first.
        
        It is very important to call this method manually during deletion as this will delete references to data objects stored in global variables of third-party modules.
        """
        if not hasattr(obj,"_modeldata"):
            return # already not initialized
        data = obj._modeldata
        
        if "_modelcache" in data:
            moddata = data["_modelcache"]
            
            if "vlists" in moddata:
                for vlist in list(moddata["vlists"].keys()):
                    del moddata["vlists"][vlist]
                del moddata["vlists"]
        
        if "_bones" in data:
            for bone in list(data["_bones"].keys()):
                del data["_bones"][bone]
            del data["_bones"]
        
        if "_anidata" in data:
            adata = data["_anidata"]
            if "_schedfunc" in adata:
                pyglet.clock.unschedule(adata["_schedfunc"])
        
        if data.get("_manual_render",False):
            del obj.batch3d
        
        # Removes open vertex lists and other caches
        
        if "_modelcache" not in data:
            return
        moddata = data["_modelcache"]
        
        if "vlist" in moddata:
            moddata["vlist"].delete() # Free up the graphics memory
        if "group" in moddata:
            del moddata["group"]
        
        del data["_modelcache"], moddata
    
    def redraw(self,obj):
        """
        Redraws the model of the given object.
        
        Note that currently this method probably won't change any data since all movement and animation is done through pyglet groups.
        """
        self.ensureModelData(obj)
        data = obj._modeldata
        
        vlists = data["_modelcache"]["vlists"]
        
        for name,region in self.modeldata["regions"].items():
            vlists[name].vertices = region.getVertices(data)
            if region.enable_tex:
                vlists[name].tex_coords = region.getTexCoords(data)
    
    def draw(self,obj):
        """
        Actually draws the model of the given object to the render target.
        
        Note that if the batch used for this object already existed, drawing will be skipped as the batch should be drawn by the owner of it.
        """
        self.ensureModelData(obj)
        
        data = obj._modeldata
        if data.get("_manual_render",False):
            obj.batch3d.draw()
    
    def remove(self,obj):
        """
        Called if the actor is removed from the world.
        
        Can be extended for custom features, currently calls :py:meth:`cleanup()`\ .
        """
        # Can be extended, but the base class should always be called
        self.cleanup(obj)
    
    def setAnimation(self,obj,animation,transition=None,force=False):
        """
        Sets the animation to be used by the object.
        
        See :py:meth:`Actor.setAnimation()` for more information.
        """
        self.ensureModelData(obj)
        data = obj._modeldata
        
        # Validity check
        if animation not in self.modeldata["animations"]:
            raise ValueError("There is no animation of name '%s' for model '%s'"%(animation,self.modelname))
        
        if data.get("_anidata",{}).get("anitype",None)==animation and not force:
            return # animation is already running
        
        # Cache the obj to improve readability
        anim = self.modeldata["animations"][animation]
        
        # Set to default if not set
        if transition is None:
            transition = anim.default_jt
        
        # Notify the animation to allow it to initialize itself
        anim.startAnimation(data,transition)
        
        # initialize animation data
        if "_anidata" not in data:
            data["_anidata"]={}
        adata = data["_anidata"]
        adata["anitype"]=animation
        
        if "_schedfunc" in adata:
            # unschedule the old animation, if any
            # prevents clashing and crashes
            pyglet.clock.unschedule(adata["_schedfunc"])
        
        # Schedule the animation function
        def schedfunc(*args):
            # This function is defined locally to create a closure
            # The closure stores the local variables, e.g. anim and data even after the parent function has finished
            # Note that this may also prevent the garbage collection of any objects defined in the parent scope
            anim.tickEntity(data)
        
        # register the function to pyglet
        pyglet.clock.schedule_interval(schedfunc,1./(anim.kps if anim.atype=="keyframes" else 60))
        # save it for later for de-initialization
        adata["_schedfunc"] = schedfunc



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["Actor","RotatableActor","RotateableActor","Controller"]

import uuid as uuidm

class Controller(object):
    """
    Base class for all controllers.
    
    Controllers define behavior of Actors and can be used to control them via e.g. the keyboard or an AI.
    
    Every controller is bound to its actor and can be enabled and disabled individually.
    You may also deactivate all controllers of an Actor by setting the ``enabled`` key of :py:attr:`Actor.controlleroptions` to False.
    """
    def __init__(self,actor):
        self.world = actor.world
        self.peng = actor.world.peng
        self.actor = actor
        self._enabled = True
        self.registerEventHandlers()
    def registerEventHandlers(self):
        """
        Method to be overridden by subclasses for registering event handlers.
        
        Automatically called upon object creation.
        """
        pass
    
    @property
    def enabled(self):
        """
        Property allowing to get and set if this controller should be active.
        
        When getting this property, the result of ANDing the internal flag and the actor flag is returned.
        
        When setting, only the local internal flag is set, allowing other controllers to still work.
        
        :raises AssertionError: when the supplied value is not of type bool
        """
        return self._enabled and self.actor.controlleroptions["enabled"]
    @enabled.setter
    def enabled(self,value):
        assert isinstance(value,bool)
        self._enabled = value

class Actor(object):
    """
    Actor object, base class for all other Actors in the world.
    
    An actor represents an object in the world, for example the player, an animal, enemy or dropped item.
    
    Everything that is not part of the terrain should be an actor.
    
    The default actor does not do anything, you should look at the subclasses for more information.
    """
    def __init__(self,peng,world,uuid=None,pos=[0,0,0]):
        self.peng = peng
        self.world = world
        self.uuid = uuid if uuid is not None else uuidm.uuid4()
        self.controllers = []
        self.controlleroptions = {
            "enabled":True
            }
        
        self._pos = pos
        
        self.model = None
    
    def addController(self,controller):
        """
        Adds a controller to the actor.
        
        A controller can control its actor and can act as a bridge between actor and user inputs.
        
        Controllers may be added anytime during the lifetime of an actor.
        """
        self.controllers.append(controller)
    
    def setModel(self,model):
        """
        Sets the model this actor should use when drawing.
        
        This method also automatically initializes the new model and removes the old, if any.
        """
        if self.model is not None:
            self.model.cleanup(self)
        self.model = model
        model.create(self)
    
    def setAnimation(self,animation,transition=None,force=False):
        """
        Sets the animation the model of this actor should show.
        
        ``animation`` is the name of the animation to switch to.
        
        ``transition`` can be used to override the transition between the animations.
        
        ``force`` can be used to force reset the animation even if it is already running.
        
        If there is no model set for this actor, a :py:exc:`RuntimeError` will be raised.
        """
        if self.model is None:
            raise RuntimeError("Can only set animation if a model is set")
        self.model.setAnimation(self,animation,transition,force)
    
    def render(self,view=None):
        """
        Called by :py:meth:`World.render3d()` to render this actor.
        
        By default, this method calls the draw method of its model, if any.
        
        For custom render behavior, it is recommended to extend this method or modify the model.
        """
        if self.model is not None:
            self.model.draw(self)
    
    # Event handlers
    def on_move(self,old):
        """
        Fake event handler called if the location of this actor changes.
        
        This handler is called after the location has changed.
        
        :param list old: The previous position
        """
        pass
    
    # Properties
    @property
    def pos(self):
        """
        Property allowing access to the position of this actor.
        
        This actor is read-write but calls :py:meth:`on_move()` if it is set.
        """
        return self._pos
    @pos.setter
    def pos(self,value):
        old = self._pos
        self._pos = value
        self.on_move(old)

class RotatableActor(Actor):
    """
    Actor that can also be rotated.
    
    This subclass adds a rotational value to the actor and a method to move the actor along the current rotation.
    """
    def __init__(self,peng,world,uuid=None,pos=[0,0,0],rot=[0,0]):
        super(RotatableActor,self).__init__(peng,world,uuid,pos)
        self._rot = rot
    def move(self,dist):
        """
        Moves the actor using standard trigonometry along the current rotational vector.
        
        :param float dist: Distance to move
        
        .. todo::
           
           Test this method, also with negative distances
        """
        x, y = self._rot
        y_angle = math.radians(y)
        x_angle = math.radians(x)
        m = math.cos(y_angle)
        dy = math.sin(y_angle)
        dy = 0.0
        dx = math.cos(x_angle)
        dz = math.sin(x_angle)
        dx,dy,dz = dx*dist,dy*dist,dz*dist
        x,y,z = self._pos
        self.pos = x+dx,y+dy,z+dz
        return dx,dy,dz
    move.__noautodoc__ = True
    
    # Event handlers
    def on_rotate(self,old):
        """
        Fake event handler called if the rotation of this actor changes.
        
        This handler is called after the rotation has been made.
        
        :param tuple old: Old rotation before rotating
        """
        pass
    
    # Properties
    @property
    def rot(self):
        """
        Property for accessing the rotation of this actor.
        
        Rotation is a tuple of ``(x,y)`` where y is clamped to -90 and 90.
        x rolls over at 360, resulting in a seamless experience for players.
        
        This property may also be written to, this calls :py:meth:`on_rotate()`\ .
        """
        return self._rot
    @rot.setter
    def rot(self,value):
        old = self._rot
        x,y = value
        y = max(-90, min(90, y))
        x %= 360
        self._rot = x,y
        self.on_rotate(old)

RotateableActor = RotatableActor # for people that spell it right



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  config.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

import sys
import os

__all__ = ["CFG_FOG_DEFAULT","CFG_LIGHT_DEFAULT","DEFAULT_CONFIG","Config"]

def _get_script_home():
    # Docs removed to not show on autodoc
    # Integrated from pyglet.resource to also work on headless distributions
    # see http://pyglet.readthedocs.io/en/pyglet-1.2-maintenance/programming_guide/resources.html
    frozen = getattr(sys, 'frozen', None)
    if frozen in ('windows_exe', 'console_exe'):
        return os.path.dirname(sys.executable)
    elif frozen == 'macosx_app':
        # py2app
        return os.environ['RESOURCEPATH']
    elif frozen:
        # pyinstaller
        return sys._MEIPASS
    else:
        main = sys.modules['__main__']
        if hasattr(main, '__file__'):
            return os.path.dirname(os.path.abspath(main.__file__))
        else:
            # cx_Freeze
            return os.path.dirname(sys.executable)

    # Probably interactive
    return os.getcwd() # Always return a path

class Config(object):
    """
    Configuration object imitating a dictionary.
    
    ``config`` can be any dictionary-style object and is used to store the configuration set by the user.
    This object only needs to implement the ``__getitem__``\ , ``__setitem__`` and ``__contains__`` special methods.
    
    ``defaults`` can be any dictionary-style object and is only read from in case the ``config`` object does not contain the key.
    Every config object is stackable, e.g. you can pass another :py:class:`Config` object as the ``defaults`` object.
    
    Example for stacking configs::
       
       >>> myconf = Config()
       >>> myconf2 = Config(defaults=myconf)
       >>> myconf["foo"] = "bar"
       >>> print(myconf2["foo"])
       bar
       >>> myconf2["bar"] = "foo"
       >>> print(myconf2["bar"])
       foo
       >>> print(myconf["bar"])
       Traceback (most recent call last):
       ...
       KeyError: Key "bar" does not exist
    
    There is no limit in stacking configurations, though higher-stacked configs may get slow when defaulting due to propagating through the whole chain.
    """
    
    def __init__(self,config=None,defaults={}):
        if config is None:
            config = {} # To avoid bugs with defaulted arguments being initialized at load time, not object instantiation.
        self.config = config
        self.defaults = defaults
    def __getitem__(self,key):
        return self.config[key] if key in self.config else self.defaults[key]
    def __setitem__(self,key,value):
        self.config[key] = value
    def __contains__(self,key):
        return key in self.config

CFG_FOG_DEFAULT = {
    "enable":False,
    
    "color":None,
    "start":128,
    "end":128+32,
    }
"""
Default fog configuration.

This configuration simply disables fog.
"""

CFG_LIGHT_DEFAULT = {"enable":False}
"""
Default lighting configuration.

This configuration simply disables lighting.
"""

DEFAULT_CONFIG = {
    # graphics.*
    # OpenGL config
    "graphics.clearColor":(0.,0.,0.,1.),
    "graphics.wireframe":False,
    "graphics.fieldofview":65.0,
    "graphics.nearclip":0.1,
    "graphics.farclip":10000, # It's over 9000!
    "graphics.min_size":None,
    "graphics.max_size":None,
    "graphics.stencil.enable":False,
    "graphics.stencil.bits":8,
    "graphics.fogSettings":Config({},defaults=CFG_FOG_DEFAULT),
    "graphics.lightSettings":Config({},defaults=CFG_LIGHT_DEFAULT),
    
    # controls.*
    # Controls
    "controls.mouse.sensitivity":0.15,
    
    "controls.controls.movespeed":10.0,
    "controls.controls.verticalspeed":5.0,
    "controls.controls.forward":"w",
    "controls.controls.backward":"s",
    "controls.controls.strafeleft":"a",
    "controls.controls.straferight":"d",
    "controls.controls.jump":"space",
    "controls.controls.crouch":"lshift",
    "controls.keybinds.strict":False,
    "controls.keybinds.debug":False,
    
    "controls.keybinds.common.copy":"Ctrl-C",
    "controls.keybinds.common.paste":"Ctrl-V",
    "controls.keybinds.common.cut":"Ctrl-X",
    
    # debug.*
    # Debug config
    "debug.events.dump":False,
    "debug.events.logerr":False,
    "debug.events.register":False,
    "debug.events.dumpfile":"",#"events.txt",
    
    # rsrc.*
    # Resource config
    "rsrc.enable":True,
    "rsrc.basepath":_get_script_home(),
    "rsrc.maxtexsize":1024, # Actual limit may be less, will be adjusted based on GL Capabilities
    
    # i18n.*
    # Translation config
    "i18n.enable":True,
    "i18n.lang":"en", # currently ISO 639-1 code, but may change in the future
    "i18n.lang.format":"{domain}:lang.{lang}", # format of resource that contains translations
    "i18n.lang.ext":".lang", # file extension including dot
    "i18n.discover_regex":r".*?/lang/(?P<lang>[a-zA-Z0-9_-]{2,5})\.lang",
    
    # events.*
    # Event config
    "events.removeonerror":True,
    "events.maxignore":3,
    
    # Other config options
    "pyglet.patch.patch_float2int":True,
}
"""
Default configuration values.

All default configuration values are stored here, for more information about specific config values, see :doc:`/configoption`\ .
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  slider.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "Progressbar","ProgressbarBackground",
    "AdvancedProgressbar",
    "Slider","SliderBackground",
    "VerticalSlider","VerticalSliderBackground",
]

import pyglet
from pyglet.gl import *

from .widgets import Background,Widget
from .button import ButtonBackground

basestring = str # for py2 compat, may get dropped at a later release

class ProgressbarBackground(Background):
    """
    Background for the :py:class:`Progressbar` Widget.
    
    This background displays a bar with a border similar to :py:class:`ButtonBackground`\ .
    Note that two colors may be given, one for the left and one for the right.
    """
    def __init__(self,widget,border,borderstyle,colors):
        super(ProgressbarBackground,self).__init__(widget)
        self.border = border
        self.borderstyle = borderstyle
        self.colors = colors
    def init_bg(self):
        self.vlist = self.submenu.batch2d.add(24,GL_QUADS,pyglet.graphics.OrderedGroup(1),
            "v2f",
            "c3B",
            )
        self.reg_vlist(self.vlist)
    def redraw_bg(self):
        x,y = self.widget.pos
        sx,sy = self.widget.size
        bx,by = self.border
        
        nmin,nmax,n = self.widget.nmin,self.widget.nmax,float(self.widget.n)
        if (nmax-nmin)==0:
            p = 0 # prevents ZeroDivisionError
        else:
            p = min((n-nmin)/(nmax-nmin),1.)
        
        # Outer vertices
        #    x          y
        v1 = x,         y+sy
        v2 = x+sx,      y+sy
        v3 = x,         y
        v4 = x+sx,      y
        
        # Inner vertices
        #    x          y
        v5 = x+bx,      y+sy-by
        v6 = x+sx-bx,   y+sy-by
        v7 = x+bx,      y+by
        v8 = x+sx-bx,   y+by

        v9 = x+(sx-bx)*p,y+by
        v10= x+(sx-bx)*p,y+sy-by
        if p<=0:
            v9,v10=v7,v5
        
        # 5 Quads, for edges and the center
        qb1 = v5+v6+v2+v1
        qb2 = v8+v4+v2+v6
        qb3 = v3+v4+v8+v7
        qb4 = v3+v7+v5+v1
        qc1 = v7+v8+v6+v5
        qc2 = v7+v9+v10+v5
        
        v = qb1+qb2+qb3+qb4+qc1+qc2
        
        self.vlist.vertices = v
        
        bg = self.submenu.bg[:3] if isinstance(self.submenu.bg,list) or isinstance(self.submenu.bg,tuple) else [242,241,240]
        o,i = bg, [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]
        s,h = [max(bg[0]-40,0),max(bg[1]-40,0),max(bg[2]-40,0)], [min(bg[0]+12,255),min(bg[1]+12,255),min(bg[2]+12,255)]
        # Outer,Inner,Shadow,Highlight
        j,k = self.colors # Other progress color
        
        if self.borderstyle == "flat":
            # Flat style makes no difference between normal,hover and pressed
            cb1 = i+i+i+i
            cb2 = i+i+i+i
            cb3 = i+i+i+i
            cb4 = i+i+i+i
            cc1 = i+i+i+i
            cc2 = j+k+k+j
        elif self.borderstyle == "gradient":
            cb1 = i+i+o+o
            cb2 = i+o+o+i
            cb3 = o+o+i+i
            cb4 = o+i+i+o
            cc1 = i+i+i+i
            cc2 = j+k+k+j
        elif self.borderstyle == "oldshadow":
            cb1 = h+h+h+h
            cb2 = s+s+s+s
            cb3 = s+s+s+s
            cb4 = h+h+h+h
            cc1 = i+i+i+i
            cc2 = j+k+k+j
        elif self.borderstyle == "material":
            cb1 = s+s+o+o
            cb2 = s+o+o+s
            cb3 = o+o+s+s
            cb4 = o+s+s+o
            cc1 = i+i+i+i
            cc2 = j+k+k+j
        else:
            raise ValueError("Invalid Border style")
        
        c = cb1+cb2+cb3+cb4+cc1+cc2
        
        self.vlist.colors = c

class Progressbar(Widget):
    """
    Progressbar displaying a progress of any action to the user.
    
    By default, this Widget uses :py:class:`ProgressbarBackground` as its Background class.
    
    The border and borderstyle options are the same as for the :py:class:`peng3d.gui.button.Button` Widget.
    
    The two colors given are for left and right, respectively. This may be used to create gradients.
    
    ``nmin``\ , ``nmax`` and ``n`` represent the minimal value, maximal value and current value, respectively.
    Unexpected behavior may occur if the minimal value is bigger then the maximum value.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 bg=None,
                 nmin=0,nmax=100,n=0,
                 border=[4,4],
                 borderstyle="flat",
                 colors=[[240,119,70],[240,119,70]],
                 ):
        self._nmin = nmin
        self._nmax = nmax
        self._n = n
        if bg is None:
            bg = ProgressbarBackground(self,border,borderstyle,colors)
        super(Progressbar,self).__init__(name,submenu,window,peng,pos,size,bg)
        self.redraw()
    
    @property
    def nmin(self):
        """
        Property representing the minimal value of the progressbar. Typically ``0``\ .
        """
        return self._nmin
    @nmin.setter
    def nmin(self,value):
        self._nmin = value
        self.redraw()
    
    @property
    def nmax(self):
        """
        Property representing the maximum value of the progressbar. Typically ``100`` to represent percentages easily.
        """
        return self._nmax
    @nmax.setter
    def nmax(self,value):
        self._nmax = value
        self.redraw()
    
    @property
    def n(self):
        """
        Property representing the current value of the progressbar.
        
        Changing this property will activate the ``progresschange`` action.
        """
        return self._n
    @n.setter
    def n(self,value):
        value = min(max(value,self.nmin),self.nmax)
        if self._n != value:
            self.doAction("progresschange")
        self._n = value
        self.redraw()
    
    @property
    def value(self):
        """
        Alias to the :py:attr:`n` property.
        """
        return self.n
    @value.setter
    def value(self,value):
        # May be a tiny bit slower, but safer if n is changed
        self.n = value

class AdvancedProgressbar(Progressbar):
    """
    Advanced Progressbar displaying the combined progress through multiple actions.
    
    Visually, this widget is identical to :py:class:`Progressbar` with the only difference
    being the way the progress percentage is calculated.
    
    The ``offset_nmin``\ , ``offset_n`` and ``offset_nmax`` parameters are equivalent
    to the parameters of the same name minus the ``offset_`` prefix.
    
    ``categories`` may be any dictionary mapping category names to 3-tuples of
    format ``(nmin,n,nmax)``\ .
    
    It is possible to read, write and delete categories through the ``widget[cat]`` syntax.
    Note however, that modifying categories in-place, e.g. like ``widget[cat][1]=100``\ , 
    requires a manual call to :py:meth:`redraw()`\ .
    
    When setting the :py:attr:`nmin`\ , :py:attr:`n` or :py:attr:`nmax` properties, only
    an internal offset value will be modified. This may result in otherwise unexpected behavior
    if setting e.g. ``n`` to ``nmax`` because the categories may influence the total percentage calculation.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None,size=None,
                 bg=None,
                 categories={},
                 offset_nmin=0,offset_nmax=0,offset_n=0,
                 border=[4,4],
                 borderstyle="flat",
                 colors=[[240,119,70],[240,119,70]],
                 ):
        super(AdvancedProgressbar,self).__init__(name,submenu,window,peng,pos,size,bg,offset_nmin,offset_nmax,offset_n,border,borderstyle,colors)
        
        self.categories = categories
        for cname,cdat in self.categories.items():
            assert len(cdat)==3 # nmin,n,nmax
    
    @property
    def nmin(self):
        return self._nmin+sum(map(lambda cdat:cdat[0],self.categories.values()))
    @nmin.setter
    def nmin(self,value):
        # may confuse users if base_nmax is very low but the categories nmax is high
        # may also prevent the expected behavior that if n is set to nmin, p is equal to 0%
        self._nmin = value
        self.redraw()
    
    @property
    def n(self):
        return self._n+sum(map(lambda cdat:cdat[1],self.categories.values()))
    @n.setter
    def n(self,value):
        # see nmin for information about some of the implications of this behavior
        self._n = value
        self.redraw()
    
    @property
    def nmax(self):
        return self._nmax+sum(map(lambda cdat:cdat[2],self.categories.values()))
    @nmax.setter
    def nmax(self,value):
        # see nmin for information about some of the implications of this behavior
        self._nmax = value
        self.redraw()
    
    def __getitem__(self,key):
        # returns the 3-tuple associated with the category
        if key not in self.categories:
            raise KeyError("No Category with name '%s'"%key)
        return self.categories[key]
        # TODO: automatically redraw if list returned here is modified
    def __setitem__(self,key,value):
        # sets the 3-tuple associated with the category
        # mostly used for category creation, since expressions of the form
        # widget[category][0]=1 will only use __getitem__ and modify data in-place
        assert isinstance(key,basestring) # py2 compat is done at the top
        assert len(value)==3 # nmin,n,nmax
        self.categories[key]=list(value) # conversion to list allows in-place modification if a tuple was passed
        self.redraw()
    def __delitem__(self,key):
        if key not in self.categories:
            raise KeyError("No Category with name '%s'"%name)
        del self.categories[key]
        self.redraw()
    
    def addCategory(self,name,nmin=0,n=0,nmax=100):
        """
        Adds a category with the given name.
        
        If the category already exists, a :py:exc:`KeyError` will be thrown. Use
        :py:meth:`updateCategory()` instead if you want to update a category.
        """
        assert isinstance(name,basestring) # py2 compat is done at the top
        if name in self.categories:
            raise KeyError("Category with name '%s' already exists"%name)
        self.categories[name]=[nmin,n,nmax]
        self.redraw()
    def updateCategory(self,name,nmin=None,n=None,nmax=None):
        """
        Smartly updates the given category.
        
        Only values that are given will be updated, others will be left unchanged.
        
        If the category does not exist, a :py:exc:`KeyError` will be thrown. Use
        :py:meth:`addCategory()` instead if you want to add a category.
        """
        # smart update, only stuff that was given
        if name not in self.categories:
            raise KeyError("No Category with name '%s'"%name)
        if nmin is not None:
            self.categories[name][0]=nmin
        if n is not None:
            self.categories[name][1]=n
        if nmax is not None:
            self.categories[name][2]=nmax
        self.redraw()
        self.doAction("progresschange")
    def deleteCategory(self,name):
        """
        Deletes the category with the given name.
        
        If the category does not exist, a :py:exc:`KeyError` will be thrown.
        """
        if name not in self.categories:
            raise KeyError("No Category with name '%s'"%name)
        del self.categories[name]
        self.redraw()

class SliderBackground(ButtonBackground):
    """
    Background for the :py:class:`Slider` Widget.
    
    This background displays a button-like handle on top of a bar representing the selectable range.
    
    All given parameters will affect the handle.
    """
    def init_bg(self):
        self.vlist_bg = self.submenu.batch2d.add(4,GL_QUADS,pyglet.graphics.OrderedGroup(0),
            "v2f",
            "c3B",
            )
        self.reg_vlist(self.vlist_bg)
        super(SliderBackground,self).init_bg()
    def redraw_bg(self):
        super(SliderBackground,self).redraw_bg()
        
        sx,sy,x,y,bx,by = super(SliderBackground,self).getPosSize()
        
        #    x          y
        v5 = x+bx,      y+sy-by
        v6 = x+sx-bx,   y+sy-by
        v7 = x+bx,      y+by
        v8 = x+sx-bx,   y+by
        
        qbg = v7+v8+v6+v5
        
        bg,_,_,_,_ = self.getColors()
        
        cbg = [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]*4
        
        self.vlist_bg.vertices = qbg
        self.vlist_bg.colors=cbg
    
    def getPosSize(self):
        sx,sy = self.widget.handlesize
        bx,by = self.border
        x,y = self.widget.pos[0]+(self.widget.size[0]-bx*2)*self.widget.p,self.widget.pos[1]
        return sx,sy,x,y,bx,by
    getPosSize.__noautodoc__ = True

class Slider(Progressbar):
    """
    Slider that can be used to get a number from the user.
    
    By default, this Widget uses :py:class:`SliderBackground` as its Background class.
    
    Most options are the same as for :py:class:`Progressbar`\ .
    
    ``handlesize`` simply determines the size of the handle.
    
    Note that scaling this widget on the y-axis will not do much, scale the handlesize instead.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[100,24],bg=None,
                 border=[4,4], borderstyle="flat",
                 nmin=0,nmax=100,n=0,
                 handlesize=[16,24],
                 ):
        self.handlesize = handlesize
        if bg is None:
            bg = SliderBackground(self,border,borderstyle)
        super(Slider,self).__init__(name,submenu,window,peng,pos,size,bg,nmin,nmax,n)
    def on_mouse_drag(self,x,y,dx,dy,button,modifiers):
        if not self.pressed:
            return
        totx = self.size[0]-self.bg.border[0]*2
        x = x-self.pos[0]-self.bg.border[0]
        x = min(max(x,0),totx)
        n = int(((x/totx)*(self.nmax-self.nmin))+.5)+self.nmin
        n = min(max(n,self.nmin),self.nmax)
        self.n = n
    
    @property
    def p(self):
        """
        Helper property containing the percentage this slider is "filled".
        
        This property is read-only.
        """
        return (self.n-self.nmin)/max((self.nmax-self.nmin),1)


class VerticalSliderBackground(SliderBackground):
    """
    Background for the :py:class:`VerticalSlider` Widget.
    
    This background uses the same technique as :py:class:`SliderBackground`\ , simply turned by 90 Degrees.
    """
    
    def getPosSize(self):
        sx,sy = self.widget.handlesize
        bx,by = self.border
        x,y = self.widget.pos[0],self.widget.pos[1]+(self.widget.size[1]-by*2)*self.widget.p
        return sx,sy,x,y,bx,by
    getPosSize.__noautodoc__ = True

class VerticalSlider(Slider):
    """
    Vertical slider that can be used as a scrollbar or getting other input.
    
    By default, this Widget uses :py:class:`VerticalSliderBackground` as its Background class.
    
    This widget is essentially the same as :py:class:`Slider`\ , only vertical.
    
    Note that you may need to flip the x and y values of ``size``\ , ``handlesize`` and ``border`` compared to :py:class:`Slider`\ .
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[24,100], bg=None,
                 border=[4,4], borderstyle="flat",
                 nmin=0,nmax=100,n=0,
                 handlesize=[24,16],
                 ):
        if bg is None:
            bg = VerticalSliderBackground(self,border,borderstyle)
        super(VerticalSlider,self).__init__(name,submenu,window,peng,pos,size,bg,border,borderstyle,nmin,nmax,n,handlesize)
    def on_mouse_drag(self,x,y,dx,dy,button,modifiers):
        if not self.pressed:
            return
        toty = self.size[1]-self.bg.border[1]*2
        y = y-self.pos[1]-self.bg.border[1]
        y = min(max(y,0),toty)
        n = int(((y/toty)*(self.nmax-self.nmin))+.5)+self.nmin
        n = min(max(n,self.nmin),self.nmax)
        self.n = n



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  camera.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = ["Camera","CameraActorFollower"]

class Camera(object):
    """
    Camera object representing a location to draw from.
    
    Each :py:class:`Camera` object is bound to a world and has three properties: a name, :py:attr:`pos` and :py:attr:`rot`\ .
    
    The name of the camera can be any string and is used to identify the camera and thus should be unique.
    """
    def __init__(self,world,name,pos=None,rot=None):
        if not (isinstance(name,str) or isinstance(name,bytes) or isinstance(unicode)):
            raise TypeError("name must be an instance of basestring!")
        self.world = world
        self._pos = pos if pos is not None else [0,0,0]
        self._rot = rot if rot is not None else [0,0]
        self.name = name
    
    def on_activate(self,old):
        """
        Fake event handler called when this camera is made current by a :py:class:`WorldView()` object.
        """
        pass
    def on_rotate(self,old,new):
        """
        Fake event handler called when this camera is rotated.
        
        The ``old`` and ``new`` parameters are both rotations and are not equal. Each parameter is a 2-tuple of ``(yaw,pitch)``\ .
        """
        pass
    def on_move(self,old,new):
        """
        Fake event handler called when this camera moves.
        
        The ``old`` and ``new`` parameters are both 3D Locations and are not equal. Each parameter is a 3-tuple of ``(x,y,z)`` in world coordinates.
        """
        pass
    
    # Move/Rotate methods
    
    @property
    def pos(self):
        """
        Property for accessing the position of the camera.
        
        This property uses a setter to call the :py:meth:`on_move()` method if set and the new location is not equal to the old location.
        """
        return self._pos
    @pos.setter
    def pos(self,value):
        if self._pos == value:
            return # Position unchanged
        old = self._pos
        self._pos = value
        self.on_move(old,value)
    
    @property
    def rot(self):
        """
        Property for accessing the rotation of the camera.
        
        This property uses a setter to call the :py:meth:`on_rotate()` method if set and the new location is not equal to the old location.
        """
        return self._rot
    @rot.setter
    def rot(self,value):
        if self._rot == value:
            return # Rotation unchanged
        old = self._rot
        self._rot = value
        self.on_rotate(old,value)

class CameraActorFollower(Camera):
    """
    Special Camera that follows the specified :py:class:`Actor()`\ .
    
    Note that neither the :py:meth:`on_move() <Camera.on_move>` nor the :py:meth:`on_rotate() <Camera.on_rotate>` event handlers are called due to the way the updating works.
    """
    def __init__(self,world,name,actor):
        super(CameraActorFollower,self).__init__(world,name)
        del self._pos,self._rot
        self.actor = actor
    
    @property
    def pos(self):
        """
        This property always equals the value of ``self.actor.pos``\ .
        
        This property may also be written to.
        """
        return self.actor.pos
    @pos.setter
    def pos(self,value):
        self.actor.pos = value
    
    @property
    def rot(self):
        """
        This property always equals the value of ``self.actor.rot``\ .
        
        This property may also be written to.
        """
        return self.actor.rot
    @rot.setter
    def rot(self,value):
        self.actor.rot = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  button.py
#  
#  Copyright 2016 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

__all__ = [
    "Button","ButtonBackground",
    "ImageButton","ImageBackground",
    "FramedImageButton","FramedImageBackground",
    "ToggleButton",
    "Checkbox","CheckboxBackground"
]

import pyglet
from pyglet.gl import *

from .widgets import Background,Widget,mouse_aabb

LABEL_FONT_SIZE = 16


class ButtonBackground(Background):
    """
    Background for the :py:class:`Button` Widget.
    
    This background renders the button and its border, but not the label.
    """
    n_vertices = 20
    change_on_press = True
    vlist_layer = 0 # used as the first argument to OrderedGroup
    
    def __init__(self,widget,border=[4,4],borderstyle="flat",
                 batch=None,change_on_press=None):
        
        super(ButtonBackground,self).__init__(widget)
        
        self.border = border
        self.borderstyle = borderstyle
        
        self.batch = batch
        
        self.change_on_press = change_on_press if change_on_press is not None else self.change_on_press
        
        self.borderstyles = {}
        self.addBorderstyle("flat",self.bs_flat)
        self.addBorderstyle("gradient",self.bs_gradient)
        self.addBorderstyle("oldshadow",self.bs_oldshadow)
        self.addBorderstyle("material",self.bs_material)
    def init_bg(self):
        # Can only be initialized here due to order of initialization
        self.batch = self.batch if self.batch is not None else self.submenu.batch2d
        
        self.vlist = self.batch.add(self.n_vertices,GL_QUADS,pyglet.graphics.OrderedGroup(self.vlist_layer),
            "v2f",
            "c3B",
            )
        self.reg_vlist(self.vlist)
    def redraw_bg(self):
        # Convenience variables
        sx,sy,x,y,bx,by = self.getPosSize()
        
        # Button background
        
        # Outer vertices
        #    x          y
        v1 = x,         y+sy
        v2 = x+sx,      y+sy
        v3 = x,         y
        v4 = x+sx,      y
        
        # Inner vertices
        #    x          y
        v5 = x+bx,      y+sy-by
        v6 = x+sx-bx,   y+sy-by
        v7 = x+bx,      y+by
        v8 = x+sx-bx,   y+by
        
        # 5 Quads, for edges and the center
        qb1 = v5+v6+v2+v1
        qb2 = v8+v4+v2+v6
        qb3 = v3+v4+v8+v7
        qb4 = v3+v7+v5+v1
        qc  = v7+v8+v6+v5
        
        v = qb1+qb2+qb3+qb4+qc
        self.vlist.vertices = v
        
        if self.borderstyle not in self.borderstyles:
            raise ValueError("Invalid Border style")
        c = self.borderstyles[self.borderstyle](*self.getColors())
        self.vlist.colors = c
    
    def getPosSize(self):
        """
        Helper function converting the actual widget position and size into a usable and offsetted form.
        
        This function should return a 6-tuple of ``(sx,sy,x,y,bx,by)`` where sx and sy are the size, x and y the position and bx and by are the border size.
        
        All values should be in pixels and already include all offsets, as they are used directly for generation of vertex data.
        
        This method can also be overridden to limit the background to a specific part of its widget.
        """
        sx,sy = self.widget.size
        x,y = self.widget.pos
        bx,by = self.border
        return sx,sy,x,y,bx,by
    def getColors(self):
        """
        Overrideable function that generates the colors to be used by various borderstyles.
        
        Should return a 5-tuple of ``(bg,o,i,s,h)``\ .
        
        ``bg`` is the base color of the background.
        
        ``o`` is the outer color, it is usually the same as the background color.
        
        ``i`` is the inner color, it is usually lighter than the background color.
        
        ``s`` is the shadow color, it is usually quite a bit darker than the background.
        
        ``h`` is the highlight color, it is usually quite a bit lighter than the background.
        """
        bg = self.submenu.bg[:3] if isinstance(self.submenu.bg,list) or isinstance(self.submenu.bg,tuple) else [242,241,240]
        o,i = bg, [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]
        s,h = [max(bg[0]-40,0),max(bg[1]-40,0),max(bg[2]-40,0)], [min(bg[0]+12,255),min(bg[1]+12,255),min(bg[2]+12,255)]
        # Outer,Inner,Shadow,Highlight
        return bg,o,i,s,h
    
    def addBorderstyle(self,name,func):
        """
        Adds a borderstyle to the background object.
        
        Note that borderstyles must be registered seperately for each background object.
        
        ``name`` is the (string) name of the borderstyle.
        
        ``func`` will be called with its arguments as ``(bg,o,i,s,h)``\ , see :py:meth:`getColors()` for more information.
        """
        self.borderstyles[name]=func
    
    @property
    def pressed(self):
        """
        Read-only helper property to be used by borderstyles for determining if the widget should be rendered as pressed or not.
        
        Note that this property may not represent the actual pressed state, it will always be False if ``change_on_press`` is disabled.
        """
        return self.change_on_press and self.widget.pressed
    @property
    def is_hovering(self):
        """
        Read-only helper property to be used by borderstyles for determining if the widget should be rendered as hovered or not.
        
        Note that this property may not represent the actual hovering state, it will always be False if ``change_on_press`` is disabled.
        """
        return self.change_on_press and self.widget.is_hovering
    
    def bs_flat(self,bg,o,i,s,h):
        # Flat style makes no difference between normal,hover and pressed
        cb1 = i+i+i+i
        cb2 = i+i+i+i
        cb3 = i+i+i+i
        cb4 = i+i+i+i
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    bs_flat.__noautodoc__ = True
    def bs_gradient(self,bg,o,i,s,h):
        if self.pressed:
            i = s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
        cb1 = i+i+o+o
        cb2 = i+o+o+i
        cb3 = o+o+i+i
        cb4 = o+i+i+o
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    bs_gradient.__noautodoc__ = True
    def bs_oldshadow(self,bg,o,i,s,h):
        if self.pressed:
            i = s
            s,h = h,s
        elif self.is_hovering:
            i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
            s = [min(s[0]+6,255),min(s[1]+6,255),min(s[2]+6,255)]
        cb1 = h+h+h+h
        cb2 = s+s+s+s
        cb3 = s+s+s+s
        cb4 = h+h+h+h
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    bs_oldshadow.__noautodoc__ = True
    def bs_material(self,bg,o,i,s,h):
        if self.pressed:
            i = [max(bg[0]-20,0),max(bg[1]-20,0),max(bg[2]-20,0)]
        elif self.is_hovering:
            i = [max(bg[0]-10,0),max(bg[1]-10,0),max(bg[2]-10,0)]
        cb1 = s+s+o+o
        cb2 = s+o+o+s
        cb3 = o+o+s+s
        cb4 = o+s+s+o
        cc  = i+i+i+i
        
        return cb1+cb2+cb3+cb4+cc
    bs_material.__noautodoc__ = True

class Button(Widget):
    """
    Button Widget allowing the user to trigger specific actions.
    
    By default, this Widget uses :py:class:`ButtonBackground` as its Background class.
    
    The border given is in pixels from the left/right and top/bottom, respectively.
    
    The borderstyle may be either ``flat``\ , which has no border at all, 
    ``gradient``\ , which fades from the inner color to the background color, 
    ``oldshadow``\ , which uses a simple fake shadow with the light from the top-left corner and
    ``material``\ , which imitates Google Material Design shadows.
    
    Also, the label of the button may only be a single line of text, anything else may produce undocumented behavior.
    
    If necessary, the font size of the Label may be changed via the global Constant :py:data:`LABEL_FONT_SIZE`\ , changes will only apply to Buttons created after change.
    The text color used is ``[62,67,73,255]`` in RGBA and the font used is Arial, which should be available on most systems.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[100,24],bg=None,
                 border=[4,4], borderstyle="flat",
                 label="Button",min_size=None):
        if bg is None:
            bg = ButtonBackground(self,border,borderstyle)
        super(Button,self).__init__(name,submenu,window,peng,pos,size,bg,min_size)
        self._label = pyglet.text.Label(str(label),
                font_name="Arial",
                font_size=LABEL_FONT_SIZE,
                color=[62,67,73,255],
                x=0,y=0,
                batch=self.submenu.batch2d,
                group=pyglet.graphics.OrderedGroup(1),
                anchor_x="center", anchor_y="center"
                )
        if getattr(label,"_dynamic",False):
            def f():
                self.label = str(label)
            self.peng.i18n.addAction("setlang",f)
        self.peng.i18n.addAction("setlang",self.redraw) # for dynamic size
        self.redraw()
        
        # Redraws the button every 2 seconds to prevent glitched graphics
        pyglet.clock.schedule_interval(lambda dt:self.redraw(),2)
    
    def on_redraw(self):
        super(Button,self).on_redraw()
        self.redraw_label()
    def redraw_label(self):
        """
        Re-draws the label by calculating its position.
        
        Currently, the label will always be centered on the Button.
        """
        # Convenience variables
        sx,sy = self.size
        x,y = self.pos
        
        # Label position
        self._label.x = x+sx/2.
        self._label.y = y+sy/2.
        self._label._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
    
    @property
    def label(self):
        """
        Property for accessing the label of this Button.
        """
        return self._label.text
    @label.setter
    def label(self,label):
        # TODO: make this work with changing languages if previous label was not dynamic
        self._label.text = str(label)
        self.redraw() # necessary for size/pos that depends on label size
    
    def getContentSize(self):
        l = [self._label.content_width,self._label.content_height]
        b = self.bg.border # TODO: make this work with borderless backgrounds
        return [l[0]+b[0]*2,l[1]+b[1]*2]
    
    def delete(self):
        self._label.delete()
        del self._label
        super(Button,self).delete()
    delete.__noautodoc__ = True


class _FakeTexture(object):
    def __init__(self,target,texid,texcoords):
        self.target = target
        self.id = texid
        self.tex_coords = texcoords
        self.anchor_x = 0
        self.anchor_y = 0

class ImageBackground(Background):
    """
    Background for the :py:class:`ImageButton` Widget.
    
    This background renders a image given based on whether the widget is pressed, hovered over or disabled.
    
    It should also be possible to use this class as a background for most other Widgets.
    """
    def __init__(self,widget,bg_idle=[GL_TEXTURE_2D,GL_TEXTURE1,[0]*12],bg_hover=None,bg_disabled=None,bg_pressed=None):
        bg = bg_idle
        self.bg_texinfo = bg
        if bg_hover is None:
            assert bg_hover[1]==self.bg_texinfo[1] # see init_bg()
            self.bg_hover=bg
        else:
            self.bg_hover=bg_hover
        if bg_disabled is None:
            assert bg_disabled[1]==self.bg_texinfo[1] # see init_bg()
            self.bg_disabled=bg
        else:
            self.bg_disabled=bg_disabled
        if bg_pressed is None:
            assert bg_pressed[1]==self.bg_texinfo[1] # see init_bg()
            self.bg_pressed=bg
        else:
            self.bg_pressed=bg_pressed
        super(ImageBackground,self).__init__(widget)
    def init_bg(self):
        # TODO: add seperate groups per active texture, in case the different images are on different textures
        self.bg_group = pyglet.graphics.TextureGroup(_FakeTexture(*self.bg_texinfo))
        self.vlist_bg = self.submenu.batch2d.add(4,GL_QUADS,self.bg_group,
            "v2f",
            ("t3f",self.bg_texinfo[2]),
            )
        self.reg_vlist(self.vlist_bg)
    def redraw_bg(self):
        # Convenience variables
        sx,sy = self.widget.size
        x,y = self.widget.pos
        
        # Button background
        
        # Outer vertices
        #    x          y
        v1 = x,         y+sy
        v2 = x+sx,      y+sy
        v3 = x,         y
        v4 = x+sx,      y
        
        q = v3+v4+v2+v1
        
        self.vlist_bg.vertices = q
        
        # Textures
        
        if not self.widget.enabled:
            self.vlist_bg.tex_coords = self.bg_disabled[2]
        elif self.widget.pressed:
            self.vlist_bg.tex_coords = self.bg_pressed[2]
        elif self.widget.is_hovering:
            self.vlist_bg.tex_coords = self.bg_hover[2]
        else:
            self.vlist_bg.tex_coords = self.bg_texinfo[2]

class ImageButton(Button):
    """
    Subclass of :py:class:`Button` using an image as a background instead.
    
    By default, this Widget uses :py:class:`ImageBackground` as its Background class.
    
    There are no changes to any other mechanics of the Button, only visually.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[100,24],bg=None,
                 label="Button",
                 bg_idle=[GL_TEXTURE_2D,GL_TEXTURE1,[0]*12],
                 bg_hover=None,
                 bg_disabled=None,
                 bg_pressed=None,
                 ):
        if bg is None:
            bg = ImageBackground(self,bg_idle,bg_hover,bg_disabled,bg_pressed)
        super(ImageButton,self).__init__(name,submenu,window,peng,pos,size,bg,label=label)


class FramedImageBackground(ImageBackground):
    """
    Background for the :py:class:`FramedImageButton` Widget.
    
    This background is similar to :py:class:`ImageBackground`\ , but it attempts to scale smarter with less artifacts.
    
    Note that this feature is currently not working properly, and will thus output a warning on the console if tried to use.
    """
    def __init__(self,widget,bg_idle=[GL_TEXTURE_2D,GL_TEXTURE1,[0]*12],bg_hover=None,bg_disabled=None,bg_pressed=None,frame_size=[[2,10,2],[2,10,2]]):
        print("Use FramedImageBackground with care, may produce graphical glitches and crashes")
        # TODO: fix this
        self.frame_size = frame_size
        self.repeat_edge=True
        self.repeat_center=True
        super(FramedImageBackground,self).__init__(widget,bg_idle,bg_hover,bg_disabled,bg_pressed)
    def init_bg(self):
        self.bg_group = pyglet.graphics.TextureGroup(_FakeTexture(*self.bg_texinfo))
        self.vlist_bg = self.submenu.batch2d.add(36,GL_QUADS,self.bg_group,
            "v2f",
            "t3f",
            )
        self.reg_vlist(self.vlist_bg)
    def redraw_bg(self):
        # Convenience variables
        sx,sy = self.widget.size
        x,y = self.widget.pos
        fx,fy = self.frame_size
        osx,osy = sum(fx),sum(fy)
        bsx = (osx/osy)*sy
        bsy = (osx/osy)*sy
        # frame-axis-left/right/center/up/down
        fxl,fxc,fxr = fx
        fyu,fyc,fyd = fy
        
        fxl,fxc,fxr = (fxl/sum(fx))*bsx, (fxc/sum(fx))*bsx, (fxr/sum(fx))*bsx
        fyu,fyc,fyd = (fyu/sum(fy))*bsy, (fyc/sum(fy))*bsy, (fyd/sum(fy))*bsy
        
        #assert fxl+fxc+fxr==sx
        #assert fyu+fyc+fyd==sy
        # To avoid confusion when supplied invalid numbers
        
        # 16 Vertices according to button_scheme.xcf
        
        v1  = x,         y+sy
        v2  = x+fxl,     y+sy
        v3  = x+sx-fxr,  y+sy
        v4  = x+sx,      y+sy
        v5  = x+sx,      y+sy-fyu
        v6  = x+sx,      y+fyd
        v7  = x+sx,      y
        v8  = x+sx-fxr,  y
        v9  = x+fxl,     y
        v10 = x,         y
        v11 = x,         y+fyd
        v12 = x,         y+sy-fyu
        v13 = x+fxl,     y+sy-fyu
        v14 = x+sx-fxr,  y+sy-fyu
        v15 = x+sx-fxr,  y+fyd
        v16 = x+fxl,     y+fyd
        
        bv1  = x,         y+bsy
        bv2  = x+fxl,     y+bsy
        bv3  = x+bsx-fxr,  y+bsy
        bv4  = x+bsx,      y+bsy
        bv5  = x+bsx,      y+bsy-fyu
        bv6  = x+bsx,      y+fyd
        bv7  = x+bsx,      y
        bv8  = x+bsx-fxr,  y
        bv9  = x+fxl,     y
        bv10 = x,         y
        bv11 = x,         y+fyd
        bv12 = x,         y+bsy-fyu
        bv13 = x+fxl,     y+bsy-fyu
        bv14 = x+bsx-fxr,  y+bsy-fyu
        bv15 = x+bsx-fxr,  y+fyd
        bv16 = x+fxl,     y+fyd
        
        qc1 = v10+v9 +v16+v11
        qc2 = v12+v13+v2 +v1 
        qc3 = v14+v5 +v4 +v3 
        qc4 = v8 +v7 +v6 +v15
        
        #qe1 = bv9 +bv8 +bv15+bv16
        #qe2 = bv11+bv16+bv13+bv12
        #qe3 = bv13+bv14+bv3 +bv2 
        #qe4 = bv15+bv6 +bv5 +bv14
        qe1 = v9 +v8 +v15+v16
        qe2 = v11+v16+v13+v12
        qe3 = v13+v14+v3 +v2 
        qe4 = v15+v6 +v5 +v14
        
        qc = v16+v15+v14+v13
        
        v = qc1+qc2+qc3+qc4+qe1+qe2+qe3+qe4+qc
        
        # Texture Coords
        if not self.widget.enabled:
            texcoords = self.bg_disabled[2]
        elif self.widget.pressed:
            texcoords = self.bg_pressed[2]
        elif self.widget.is_hovering:
            texcoords = self.bg_hover[2]
        else:
            texcoords = self.bg_texinfo[2]
        print("###")
        print(self.widget.enabled,self.widget.pressed,self.widget.is_hovering)
        print(texcoords)
        x,y,_,sx,_,_,_,sy,_,_,_,_ = texcoords
        sx,sy=sx-x,sy-y
        asx,asy = min(sx,(osx/osy)*sx),min(sy,(osx/osy)*sx)
        fxl,fxc,fxr=(fx[0]/sum(fx))*sx, (fx[1]/sum(fx))*sx, (fx[2]/sum(fx))*sx
        fyu,fyc,fyd=(fy[0]/sum(fy))*sy, (fy[1]/sum(fy))*sy, (fy[2]/sum(fy))*sy
        
        print(x,y,sx,sy)
        print(fxl,fxc,fxr)
        print(fyu,fyc,fyd)
        print("###")
        
        assert fxl+fxc+fxr==sx
        assert fyu+fyc+fyd==sy
        
        t1  = x,         y+sy,      0
        t2  = x+fxl,     y+sy,      0
        t3  = x+sx-fxr,  y+sy,      0
        t4  = x+sx,      y+sy,      0
        t5  = x+sx,      y+sy-fyu,  0
        t6  = x+sx,      y+fyd,     0
        t7  = x+sx,      y,         0
        t8  = x+sx-fxr,  y,         0
        t9  = x+fxl,     y,         0
        t10 = x,         y,         0
        t11 = x,         y+fyd,     0
        t12 = x,         y+sy-fyu,  0
        t13 = x+fxl,     y+sy-fyu,  0
        t14 = x+sx-fxr,  y+sy-fyu,  0
        t15 = x+sx-fxr,  y+fyd,     0
        t16 = x+fxl,     y+fyd,     0
        
        at1  = x,         y+asy,      0
        at2  = x+fxl,     y+asy,      0
        at3  = x+asx-fxr,  y+asy,      0
        at4  = x+asx,      y+asy,      0
        at5  = x+asx,      y+asy-fyu,  0
        at6  = x+asx,      y+fyd,     0
        at7  = x+asx,      y,         0
        at8  = x+asx-fxr,  y,         0
        at9  = x+fxl,     y,         0
        at10 = x,         y,         0
        at11 = x,         y+fyd,     0
        at12 = x,         y+asy-fyu,  0
        at13 = x+fxl,     y+asy-fyu,  0
        at14 = x+asx-fxr,  y+asy-fyu,  0
        at15 = x+asx-fxr,  y+fyd,     0
        at16 = x+fxl,     y+fyd,     0
        
        
        
        tqc1 = t10+t9 +t16+t11
        tqc2 = t12+t13+t2 +t1 
        tqc3 = t14+t5 +t4 +t3 
        tqc4 = t8 +t7 +t6 +t15
        
        # size of center x size adjusted
        #fxcr = (fx[1]/sum(fx))*bsx
        # rel x pos of v3 in P
        #xv3 = self.widget.size[0]-(self.widget.size[0]*((fx[2]/sum(fx))))
        xv3 = v3[0]-self.widget.pos[0]
        xv2 = v2[0]-self.widget.pos[0]
        #xv2 = self.widget.size[0]*(fx[0]/sum(fx))
        #txr = (fxcr/self.widget.size[0])*(self.widget.size[0]+((fx[2]/sum(fx))*bsx))
        #  P   Ratio of (center-x-size adjusted) to x-size applied to (x-size plus (right-x-size adjusted))
        #txr = fxl+(self.widget.size[0]*((fx[2]/sum(fx))))
        #  T   Left-x-size not adjusted plus x-size times (ratio of right-x-size to total)
        #txr = sx-fxr # Default
        #  T   x-size minus right-x-size not adjusted
        #txr = (max((xv3-xv2),fx[1])/self.widget.size[0])*sx
        print(fx[1]/sum(fx))
        txr = (fx[1]/sum(fx))*asx
        #  P  ratio of (xv3 rel to xv2) to x-size
        
        
        txr = x+txr
        
        xt2  = x+fxl,     y+asy,      0
        xt13 = x+fxl,     y+asy-fyu,  0
        xt16 = x+fxl,     y+fyd,     0
        xt9  = x+fxl,     y,         0
        
        xt8  = txr,  y,         0
        xt15 = txr,  y+fyd,     0
        xt14 = txr,  y+asy-fyu,  0
        xt3  = txr,  y+asy,      0
        
        
        yt6  = x+asx,      y+fyd,     0
        yt15 = x+asx-fxr,  y+fyd,     0
        yt11 = x,         y+fyd,     0
        yt16 = x+fxl,     y+fyd,     0
        
        yt12 = x,         y+asy-fyu,  0
        yt13 = x+fxl,     y+asy-fyu,  0
        yt5  = x+asx,      y+asy-fyu,  0
        yt14 = x+asx-fxr,  y+asy-fyu,  0
        
        tqe1 = xt9 +xt8 +xt15+xt16
        tqe2 = yt11+yt16+yt13+yt12
        tqe3 = xt13+xt14+xt3 +xt2 
        tqe4 = yt15+yt6 +yt5 +yt14
        
        #tqe1 = at9 +at8 +at15+at16
        #tqe2 = at11+at16+at13+at12
        #tqe3 = at13+at14+at3 +at2 
        #tqe4 = at15+at6 +at5 +at14
        
        #tqe1 = t9 +t8 +t15+t16
        #tqe2 = t11+t16+t13+t12
        #tqe3 = t13+t14+t3 +t2 
        #tqe4 = t15+t6 +t5 +t14
        
        tqc = t16+t15+t14+t13
        
        t = tqc1+tqc2+tqc3+tqc4+tqe1+tqe2+tqe3+tqe4+tqc
        #print(t)
        #print(v)
        self.vlist_bg.vertices=v
        self.vlist_bg.tex_coords=t

class FramedImageButton(ImageButton):
    """
    Subclass of :py:class:`ImageButton` adding smart scaling to the background.
    
    By default, this Widget uses :py:class:`FramedImageBackground` as its Background class.
    
    Note that this feature is currently not working properly, and will thus output a warning on the console if tried to use.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[100,24],bg=None,
                 label="Button",
                 bg_idle=[GL_TEXTURE_2D,GL_TEXTURE1,[0]*12],
                 bg_hover=None,
                 bg_disabled=None,
                 bg_pressed=None,
                 frame_size=[[2,10,2],[2,10,2]],
                 ):
        if bg is None:
            bg = FramedImageBackground(self,bg_idle,bg_hover,bg_disabled,bg_pressed,frame_size)
        super(FramedImageButton,self).__init__(name,submenu,window,peng,pos,size,bg,label=label)


class ToggleButton(Button):
    """
    Variant of :py:class:`Button` that stays pressed until clicked again.
    
    This widgets adds the following actions:
    
    - ``press_down`` is called upon depressing the button
    - ``press_up`` is called upon releasing the button
    - ``click`` is changed to be called on every click on the button, e.g. like ``press_down`` and ``press_up`` combined
    """
    def on_mouse_press(self,x,y,button,modifiers):
        if not self.clickable:
            return
        elif mouse_aabb([x,y],self.size,self.pos):
            if button == pyglet.window.mouse.LEFT:
                self.doAction("click")
                self.pressed = not self.pressed
                if self.pressed:
                    self.doAction("press_down")
                else:
                    self.doAction("press_up")
            elif button == pyglet.window.mouse.RIGHT:
                self.doAction("context")
            self.redraw()
    def on_mouse_release(self,x,y,button,modifiers):
        pass


class CheckboxBackground(ButtonBackground):
    """
    Background for the :py:class:`Checkbox` Widget.
    
    This background looks like a button, but adds a square in the middle if it is pressed.
    
    The color of the square defaults to a tone of orange commonly found in GTK GUIs on Ubuntu.
    """
    vlist_layer = 1
    
    def __init__(self,widget,borderstyle,checkcolor=[240,119,70],**kwargs):
        self.checkcolor = checkcolor
        super(CheckboxBackground,self).__init__(widget,[3,3],borderstyle,**kwargs)
    def init_bg(self):
        super(CheckboxBackground,self).init_bg()
        self.vlist_check = self.submenu.batch2d.add(4,GL_QUADS,pyglet.graphics.OrderedGroup(10),
            "v2f",
            "c3B",
            )
        self.reg_vlist(self.vlist_check)
    def redraw_bg(self):
        # Convenience variables
        sx,sy = self.widget.size
        x,y = self.widget.pos
        bx,by = self.border
        
        # Button background
        
        # Outer vertices
        #    x          y
        v1 = x,         y+sy
        v2 = x+sx,      y+sy
        v3 = x,         y
        v4 = x+sx,      y
        
        # Inner vertices
        #    x          y
        v5 = x+bx,      y+sy-by
        v6 = x+sx-bx,   y+sy-by
        v7 = x+bx,      y+by
        v8 = x+sx-bx,   y+by
        
        # 5 Quads, for edges and the center
        qb1 = v5+v6+v2+v1
        qb2 = v8+v4+v2+v6
        qb3 = v3+v4+v8+v7
        qb4 = v3+v7+v5+v1
        qc  = v7+v8+v6+v5
        
        v = qb1+qb2+qb3+qb4+qc
        
        self.vlist.vertices = v
        
        bg = self.submenu.bg[:3] if isinstance(self.submenu.bg,list) or isinstance(self.submenu.bg,tuple) else [242,241,240]
        o,i = bg, [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]
        s,h = [max(bg[0]-40,0),max(bg[1]-40,0),max(bg[2]-40,0)], [min(bg[0]+12,255),min(bg[1]+12,255),min(bg[2]+12,255)]
        # Outer,Inner,Shadow,Highlight
        
        if self.borderstyle == "flat":
            if self.widget.pressed:
                i = s
            cb1 = i+i+i+i
            cb2 = i+i+i+i
            cb3 = i+i+i+i
            cb4 = i+i+i+i
            cc  = i+i+i+i
        elif self.borderstyle == "gradient":
            if self.widget.pressed:
                i = s
            elif self.widget.is_hovering:
                i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
            cb1 = i+i+o+o
            cb2 = i+o+o+i
            cb3 = o+o+i+i
            cb4 = o+i+i+o
            cc  = i+i+i+i
        elif self.borderstyle == "oldshadow":
            if self.widget.pressed:
                i = s
                s,h = h,s
            elif self.widget.is_hovering:
                i = [min(i[0]+6,255),min(i[1]+6,255),min(i[2]+6,255)]
                s = [min(s[0]+6,255),min(s[1]+6,255),min(s[2]+6,255)]
            cb1 = h+h+h+h
            cb2 = s+s+s+s
            cb3 = s+s+s+s
            cb4 = h+h+h+h
            cc  = i+i+i+i
        elif self.borderstyle == "material":
            if self.widget.pressed:
                i = [max(bg[0]-20,0),max(bg[1]-20,0),max(bg[2]-20,0)]
            elif self.widget.is_hovering:
                i = [max(bg[0]-10,0),max(bg[1]-10,0),max(bg[2]-10,0)]
            cb1 = s+s+o+o
            cb2 = s+o+o+s
            cb3 = o+o+s+s
            cb4 = o+s+s+o
            cc  = i+i+i+i
        else:
            raise ValueError("Invalid Border style")
        
        c = cb1+cb2+cb3+cb4+cc
        
        self.vlist.colors = c
        
        # Cross
        
        # Old method that displayed a tick
        """if not self.widget.pressed:
            self.vlist_cross.colors = 6*bg
        else:
            if self.borderstyle=="flat":
                c = [min(bg[0]+8,255),min(bg[1]+8,255),min(bg[2]+8,255)]
            elif self.borderstyle=="gradient":
                c = h
            elif self.borderstyle=="oldshadow":
                c = h
            elif self.borderstyle=="material":
                c = s
            self.vlist_cross.colors = 6*c
        
        # Convenience variables
        sx,sy = self.widget.size
        x,y = self.widget.pos
        bx,by = self.border
        
        v1 = x+bx,      y+(sy-by*2)/2+by
        v2 = x+sx/2,    y+(sy-by*2)/4+by
        v3 = v6
        v4 = x+sx,      y+sy
        v5 = x+sx/2,    y+by
        v6 = x+bx,      y+(sy-by*2)/4+by
        
        self.vlist_cross.vertices = v2+v1+v6+v5+v4+v3"""
        
        # TODO: add better visual indicator
        
        v1 = x+bx*1.5,    y+sy-by*1.5
        v2 = x+sx-bx*1.5, y+sy-by*1.5
        v3 = x+bx*1.5,    y+by*1.5
        v4 = x+sx-bx*1.5, y+by*1.5
        
        self.vlist_check.colors = self.checkcolor*4 if self.widget.pressed else i*4
        self.vlist_check.vertices = v3+v4+v2+v1

class Checkbox(ToggleButton):
    """
    Variant of :py:class:`ToggleButton` using a different visual indicator.
    
    By default, this Widget uses :py:class:`CheckboxBackground` as its Background class.
    
    Note that the position and size given are for the indicator, the label will be bigger than the given size.
    
    The label given will be displayed to the right of the Checkbox.
    """
    def __init__(self,name,submenu,window,peng,
                 pos=None, size=[100,24],bg=None,
                 borderstyle="flat",
                 label="Checkbox",
                 checkcolor=[240,119,70]):
        if bg is None:
            bg = CheckboxBackground(self,borderstyle,checkcolor)
        super(Checkbox,self).__init__(name,submenu,window,peng,pos,size,bg,borderstyle=borderstyle,label=label)
    def redraw_label(self):
        """
        Re-calculates the position of the Label.
        """
        # Convenience variables
        sx,sy = self.size
        x,y = self.pos
        
        # Label position
        self._label.anchor_x = "left"
        self._label.x = x+sx/2.+sx
        self._label.y = y+sy/2.+sy*.15
        self._label._update() # Needed to prevent the label from drifting to the top-left after resizing by odd amounts
